# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Convert Tink Keyset with JWT keys from and to JWK sets."""

import json
import random

from typing import cast, Dict, List, Optional, Union

from tink.proto import jwt_ecdsa_pb2
from tink.proto import jwt_rsa_ssa_pkcs1_pb2
from tink.proto import jwt_rsa_ssa_pss_pb2
from tink.proto import tink_pb2
import tink
from tink.jwt import _jwt_format

_JWT_ECDSA_PUBLIC_KEY_TYPE = (
    'type.googleapis.com/google.crypto.tink.JwtEcdsaPublicKey')
_JWT_RSA_SSA_PKCS1_PUBLIC_KEY_TYPE = (
    'type.googleapis.com/google.crypto.tink.JwtRsaSsaPkcs1PublicKey')
_JWT_RSA_SSA_PSS_PUBLIC_KEY_TYPE = (
    'type.googleapis.com/google.crypto.tink.JwtRsaSsaPssPublicKey')

_ECDSA_PARAMS = {
    jwt_ecdsa_pb2.ES256: ('ES256', 'P-256'),
    jwt_ecdsa_pb2.ES384: ('ES384', 'P-384'),
    jwt_ecdsa_pb2.ES512: ('ES512', 'P-521')
}

_ECDSA_NAME_TO_ALGORITHM = {
    alg_name: algorithm for algorithm, (alg_name, _) in _ECDSA_PARAMS.items()
}

_RSA_SSA_PKCS1_PARAMS = {
    jwt_rsa_ssa_pkcs1_pb2.RS256: 'RS256',
    jwt_rsa_ssa_pkcs1_pb2.RS384: 'RS384',
    jwt_rsa_ssa_pkcs1_pb2.RS512: 'RS512'
}

_RSA_SSA_PKCS1_NAME_TO_ALGORITHM = {
    alg_name: algorithm
    for algorithm, alg_name in _RSA_SSA_PKCS1_PARAMS.items()
}

_RSA_SSA_PSS_PARAMS = {
    jwt_rsa_ssa_pss_pb2.PS256: 'PS256',
    jwt_rsa_ssa_pss_pb2.PS384: 'PS384',
    jwt_rsa_ssa_pss_pb2.PS512: 'PS512'
}

_RSA_SSA_PSS_NAME_TO_ALGORITHM = {
    alg_name: algorithm
    for algorithm, alg_name in _RSA_SSA_PSS_PARAMS.items()
}


def _base64_encode(data: bytes) -> str:
  return _jwt_format.base64_encode(data).decode('utf8')


def _base64_decode(data: str) -> bytes:
  return _jwt_format.base64_decode(data.encode('utf8'))


def from_public_keyset_handle(keyset_handle: tink.KeysetHandle) -> str:
  """Converts a Tink KeysetHandle with JWT keys into a Json Web Key (JWK) set.

  JWK is defined in https://www.rfc-editor.org/rfc/rfc7517.txt.

  Disabled keys are skipped.

  Keys with output prefix type "TINK" will include the encoded key ID as "kid"
  value. Keys with output prefix type "RAW" will not have a "kid" value set.

  Currently, public keys for algorithms ES256, ES384, ES512, RS256, RS384,
  RS512, PS256, PS384 and PS512 supported.

  Args:
    keyset_handle: A Tink KeysetHandle that contains JWT Keys.

  Returns:
    A JWK set, which is a JSON encoded string.

  Raises:
    TinkError if the keys are not of the expected type, or if they have a
    ouput prefix type that is not supported.
  """
  serialization = tink.proto_keyset_format.serialize_without_secret(
      keyset_handle
  )
  keyset = tink_pb2.Keyset.FromString(serialization)
  keys = []
  for key in keyset.key:
    if key.status != tink_pb2.ENABLED:
      continue
    if key.key_data.key_material_type != tink_pb2.KeyData.ASYMMETRIC_PUBLIC:
      raise tink.TinkError('wrong key material type')
    if key.output_prefix_type not in [tink_pb2.RAW, tink_pb2.TINK]:
      raise tink.TinkError('unsupported output prefix type')
    if key.key_data.type_url == _JWT_ECDSA_PUBLIC_KEY_TYPE:
      keys.append(_convert_jwt_ecdsa_key(key))
    elif key.key_data.type_url == _JWT_RSA_SSA_PKCS1_PUBLIC_KEY_TYPE:
      keys.append(_convert_jwt_rsa_ssa_pkcs1_key(key))
    elif key.key_data.type_url == _JWT_RSA_SSA_PSS_PUBLIC_KEY_TYPE:
      keys.append(_convert_jwt_rsa_ssa_pss_key(key))
    else:
      raise tink.TinkError('unknown key type: %s' % key.key_data.type_url)
  return json.dumps({'keys': keys}, separators=(',', ':'))


# Deprecated. Use from_public_keyset_handle instead.
def from_keyset_handle(keyset_handle: tink.KeysetHandle,
                       key_access: Optional[tink.KeyAccess] = None) -> str:
  _ = key_access
  return from_public_keyset_handle(keyset_handle)


def _convert_jwt_ecdsa_key(
    key: tink_pb2.Keyset.Key) -> Dict[str, Union[str, List[str]]]:
  """Converts a JwtEcdsaPublicKey into a JWK."""
  ecdsa_public_key = jwt_ecdsa_pb2.JwtEcdsaPublicKey.FromString(
      key.key_data.value)
  if ecdsa_public_key.algorithm not in _ECDSA_PARAMS:
    raise tink.TinkError('unknown ecdsa algorithm')
  alg, crv = _ECDSA_PARAMS[ecdsa_public_key.algorithm]
  output = {
      'kty': 'EC',
      'crv': crv,
      'x': _base64_encode(ecdsa_public_key.x),
      'y': _base64_encode(ecdsa_public_key.y),
      'use': 'sig',
      'alg': alg,
      'key_ops': ['verify'],
  }
  kid = _jwt_format.get_kid(key.key_id, key.output_prefix_type)
  if kid:
    output['kid'] = kid
  elif ecdsa_public_key.HasField('custom_kid'):
    output['kid'] = ecdsa_public_key.custom_kid.value
  return output


def _convert_jwt_rsa_ssa_pkcs1_key(
    key: tink_pb2.Keyset.Key) -> Dict[str, Union[str, List[str]]]:
  """Converts a JwtRsaSsaPkcs1PublicKey into a JWK."""
  public_key = jwt_rsa_ssa_pkcs1_pb2.JwtRsaSsaPkcs1PublicKey.FromString(
      key.key_data.value)
  if public_key.algorithm not in _RSA_SSA_PKCS1_PARAMS:
    raise tink.TinkError('unknown RSA SSA PKCS1 algorithm')
  alg = _RSA_SSA_PKCS1_PARAMS[public_key.algorithm]
  output = {
      'kty': 'RSA',
      'n': _base64_encode(public_key.n),
      'e': _base64_encode(public_key.e),
      'use': 'sig',
      'alg': alg,
      'key_ops': ['verify'],
  }
  kid = _jwt_format.get_kid(key.key_id, key.output_prefix_type)
  if kid:
    output['kid'] = kid
  elif public_key.HasField('custom_kid'):
    output['kid'] = public_key.custom_kid.value
  return output


def _convert_jwt_rsa_ssa_pss_key(
    key: tink_pb2.Keyset.Key) -> Dict[str, Union[str, List[str]]]:
  """Converts a JwtRsaSsaPssPublicKey into a JWK."""
  public_key = jwt_rsa_ssa_pss_pb2.JwtRsaSsaPssPublicKey.FromString(
      key.key_data.value)
  if public_key.algorithm not in _RSA_SSA_PSS_PARAMS:
    raise tink.TinkError('unknown RSA SSA PSS algorithm')
  alg = _RSA_SSA_PSS_PARAMS[public_key.algorithm]
  output = {
      'kty': 'RSA',
      'n': _base64_encode(public_key.n),
      'e': _base64_encode(public_key.e),
      'use': 'sig',
      'alg': alg,
      'key_ops': ['verify'],
  }
  kid = _jwt_format.get_kid(key.key_id, key.output_prefix_type)
  if kid:
    output['kid'] = kid
  elif public_key.HasField('custom_kid'):
    output['kid'] = public_key.custom_kid.value
  return output


def _generate_unused_key_id(keyset: tink_pb2.Keyset) -> int:
  while True:
    key_id = random.randint(1, 2147483647)
    if key_id not in {key.key_id for key in keyset.key}:
      return key_id


def to_public_keyset_handle(jwk_set: str) -> tink.KeysetHandle:
  """Converts a Json Web Key (JWK) set into a Tink KeysetHandle with JWT keys.

  JWK is defined in https://www.rfc-editor.org/rfc/rfc7517.txt.

  All keys are converted into Tink keys with output prefix type "RAW".

  Currently, public keys for algorithms ES256, ES384, ES512, RS256, RS384,
  RS512, PS256, PS384 and PS512 supported.

  Args:
    jwk_set: A JWK set, which is a JSON encoded string.

  Returns:
    A tink.KeysetHandle.

  Raises:
    TinkError if the key cannot be converted.
  """
  try:
    keys_dict = json.loads(jwk_set)
  except json.decoder.JSONDecodeError as e:
    raise tink.TinkError('error parsing JWK set: %s' % e.msg)
  if 'keys' not in keys_dict:
    raise tink.TinkError('invalid JWK set: keys not found')
  proto_keyset = tink_pb2.Keyset()
  for key in keys_dict['keys']:
    if 'alg' not in key:
      raise tink.TinkError('invalid JWK: alg not found')
    alg = key['alg']
    if alg.startswith('ES'):
      proto_key = _convert_to_ecdsa_key(key)
    elif alg.startswith('RS'):
      proto_key = _convert_to_rsa_ssa_pkcs1_key(key)
    elif alg.startswith('PS'):
      proto_key = _convert_to_rsa_ssa_pss_key(key)
    else:
      raise tink.TinkError('unknown alg')
    new_id = _generate_unused_key_id(proto_keyset)
    proto_key.key_id = new_id
    proto_keyset.key.append(proto_key)
    # JWK sets do not really have a primary key (see RFC 7517, Section 5.1).
    # To verify signature, it also does not matter which key is primary. We
    # simply set it to the last key.
    proto_keyset.primary_key_id = new_id
  return tink.proto_keyset_format.parse_without_secret(
      proto_keyset.SerializeToString()
  )


# Deprecated. Use to_public_keyset_handle instead.
def to_keyset_handle(
    jwk_set: str,
    key_access: Optional[tink.KeyAccess] = None) -> tink.KeysetHandle:
  _ = key_access
  return to_public_keyset_handle(jwk_set)


def _validate_use_and_key_ops(key: Dict[str, Union[str, List[str]]]):
  """Checks that 'key_ops' and 'use' have the right values if present."""
  if 'key_ops' in key:
    key_ops = key['key_ops']
    if len(key_ops) != 1 or key_ops[0] != 'verify':
      raise tink.TinkError('invalid key_ops')
  if 'use' in key and key['use'] != 'sig':
    raise tink.TinkError('invalid use')


def _convert_to_ecdsa_key(
    key: Dict[str, Union[str, List[str]]]) -> tink_pb2.Keyset.Key:
  """Converts a EC Json Web Key (JWK) into a tink_pb2.Keyset.Key."""
  ecdsa_public_key = jwt_ecdsa_pb2.JwtEcdsaPublicKey()
  algorithm = _ECDSA_NAME_TO_ALGORITHM.get(cast(str, key['alg']), None)
  if not algorithm:
    raise tink.TinkError('unknown ECDSA algorithm')
  if key.get('kty', None) != 'EC':
    raise tink.TinkError('invalid kty')
  _, crv = _ECDSA_PARAMS[algorithm]
  if key.get('crv', None) != crv:
    raise tink.TinkError('invalid crv')
  _validate_use_and_key_ops(key)
  if 'd' in key:
    raise tink.TinkError('cannot convert private ECDSA key')
  ecdsa_public_key.algorithm = algorithm
  ecdsa_public_key.x = _base64_decode(cast(str, key['x']))
  ecdsa_public_key.y = _base64_decode(cast(str, key['y']))
  if 'kid' in key:
    ecdsa_public_key.custom_kid.value = key['kid']
  proto_key = tink_pb2.Keyset.Key()
  proto_key.key_data.type_url = _JWT_ECDSA_PUBLIC_KEY_TYPE
  proto_key.key_data.value = ecdsa_public_key.SerializeToString()
  proto_key.key_data.key_material_type = tink_pb2.KeyData.ASYMMETRIC_PUBLIC
  proto_key.output_prefix_type = tink_pb2.RAW
  proto_key.status = tink_pb2.ENABLED
  return proto_key


def _convert_to_rsa_ssa_pkcs1_key(
    key: Dict[str, Union[str, List[str]]]) -> tink_pb2.Keyset.Key:
  """Converts a JWK into a JwtEcdsaPublicKey."""
  public_key = jwt_rsa_ssa_pkcs1_pb2.JwtRsaSsaPkcs1PublicKey()
  algorithm = _RSA_SSA_PKCS1_NAME_TO_ALGORITHM.get(cast(str, key['alg']), None)
  if not algorithm:
    raise tink.TinkError('unknown RSA SSA PKCS1 algorithm')
  if key.get('kty', None) != 'RSA':
    raise tink.TinkError('invalid kty')
  _validate_use_and_key_ops(key)
  if ('p' in key or 'q' in key or 'dp' in key or 'dq' in key or 'd' in key or
      'qi' in key):
    raise tink.TinkError('importing RSA private keys is not implemented')
  public_key.algorithm = algorithm
  public_key.n = _base64_decode(cast(str, key['n']))
  public_key.e = _base64_decode(cast(str, key['e']))
  if 'kid' in key:
    public_key.custom_kid.value = key['kid']
  proto_key = tink_pb2.Keyset.Key()
  proto_key.key_data.type_url = _JWT_RSA_SSA_PKCS1_PUBLIC_KEY_TYPE
  proto_key.key_data.value = public_key.SerializeToString()
  proto_key.key_data.key_material_type = tink_pb2.KeyData.ASYMMETRIC_PUBLIC
  proto_key.output_prefix_type = tink_pb2.RAW
  proto_key.status = tink_pb2.ENABLED
  return proto_key


def _convert_to_rsa_ssa_pss_key(
    key: Dict[str, Union[str, List[str]]]) -> tink_pb2.Keyset.Key:
  """Converts a JWK into a JwtEcdsaPublicKey."""
  public_key = jwt_rsa_ssa_pss_pb2.JwtRsaSsaPssPublicKey()
  algorithm = _RSA_SSA_PSS_NAME_TO_ALGORITHM.get(cast(str, key['alg']), None)
  if not algorithm:
    raise tink.TinkError('unknown RSA SSA PSS algorithm')
  if key.get('kty', None) != 'RSA':
    raise tink.TinkError('invalid kty')
  _validate_use_and_key_ops(key)
  if ('p' in key or 'q' in key or 'dp' in key or 'dq' in key or 'd' in key or
      'qi' in key):
    raise tink.TinkError('importing RSA private keys is not implemented')
  public_key.algorithm = algorithm
  public_key.n = _base64_decode(cast(str, key['n']))
  public_key.e = _base64_decode(cast(str, key['e']))
  if 'kid' in key:
    public_key.custom_kid.value = cast(str, key['kid'])
  proto_key = tink_pb2.Keyset.Key()
  proto_key.key_data.type_url = _JWT_RSA_SSA_PSS_PUBLIC_KEY_TYPE
  proto_key.key_data.value = public_key.SerializeToString()
  proto_key.key_data.key_material_type = tink_pb2.KeyData.ASYMMETRIC_PUBLIC
  proto_key.output_prefix_type = tink_pb2.RAW
  proto_key.status = tink_pb2.ENABLED
  return proto_key



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tink.python.tink.json_json_proto_keyset_format."""

import io

from absl.testing import absltest

import tink
from tink import aead
from tink import core
from tink import hybrid
from tink import secret_key_access
from tink import tink_config


def setUpModule():
  tink_config.register()


class InvalidKeyAccess(core.KeyAccess):
  pass


class TinkJsonProtoKeysetFormatTest(absltest.TestCase):

  def test_serialize_parse(self):
    keyset_handle = tink.new_keyset_handle(aead.aead_key_templates.AES128_GCM)
    serialized_keyset = tink.json_proto_keyset_format.serialize(
        keyset_handle, secret_key_access.TOKEN
    )
    parsed_keyset_handle = tink.json_proto_keyset_format.parse(
        serialized_keyset, secret_key_access.TOKEN
    )

    # check that keyset_handle and parsed_handle are the same.
    plaintext = b'plaintext'
    associated_data = b'associated_data'
    primitive1 = keyset_handle.primitive(aead.Aead)
    ciphertext = primitive1.encrypt(plaintext, associated_data)
    primitive2 = parsed_keyset_handle.primitive(aead.Aead)
    self.assertEqual(primitive2.decrypt(ciphertext, associated_data), plaintext)

    # check that serialize and parse fail without secret_key_access.TOKEN
    with self.assertRaises(core.TinkError):
      tink.json_proto_keyset_format.serialize(keyset_handle, InvalidKeyAccess())
    with self.assertRaises(core.TinkError):
      tink.json_proto_keyset_format.parse(serialized_keyset, InvalidKeyAccess())

  def test_serialize_parse_without_secret(self):
    private_handle = tink.new_keyset_handle(
        hybrid.hybrid_key_templates.ECIES_P256_HKDF_HMAC_SHA256_AES128_GCM
    )
    public_handle = private_handle.public_keyset_handle()

    # serialize and parse public_handle
    serialized = tink.json_proto_keyset_format.serialize_without_secret(
        public_handle
    )
    parsed = tink.json_proto_keyset_format.parse_without_secret(serialized)

    # check that parsed works with private_handle
    plaintext = b'plaintext'
    context_info = b'context info'
    hybrid_enc = parsed.primitive(hybrid.HybridEncrypt)
    ciphertext = hybrid_enc.encrypt(plaintext, context_info)
    hybrid_dec = private_handle.primitive(hybrid.HybridDecrypt)
    self.assertEqual(hybrid_dec.decrypt(ciphertext, context_info), plaintext)

  def test_serialize_parse_encrypted(self):
    keyset_encryption_aead = tink.new_keyset_handle(
        aead.aead_key_templates.AES128_GCM
    ).primitive(aead.Aead)

    keyset_encryption_associated_data = b'keyset_encryption_associated_data'
    keyset_handle = tink.new_keyset_handle(aead.aead_key_templates.AES128_GCM)
    encrypted_keyset = tink.json_proto_keyset_format.serialize_encrypted(
        keyset_handle, keyset_encryption_aead, keyset_encryption_associated_data
    )
    parsed_keyset_handle = tink.json_proto_keyset_format.parse_encrypted(
        encrypted_keyset,
        keyset_encryption_aead,
        keyset_encryption_associated_data,
    )

    # check that keyset_handle and parse_handle are the same.
    plaintext = b'plaintext'
    associated_data = b'associated_data'
    primitive1 = keyset_handle.primitive(aead.Aead)
    ciphertext = primitive1.encrypt(plaintext, associated_data)
    primitive2 = parsed_keyset_handle.primitive(aead.Aead)
    self.assertEqual(primitive2.decrypt(ciphertext, associated_data), plaintext)

    # check that it serialize and parse fail without secret_key_access.TOKEN
    with self.assertRaises(core.TinkError):
      tink.json_proto_keyset_format.parse_encrypted(
          encrypted_keyset,
          keyset_encryption_aead,
          b'invalid_associated_data',
      )

  def test_serialize_encrypted_read_keyset_handle_with_associated_data(self):
    keyset_encryption_aead = tink.new_keyset_handle(
        aead.aead_key_templates.AES128_GCM
    ).primitive(aead.Aead)

    keyset_encryption_associated_data = b'keyset_encryption_associated_data'
    keyset_handle = tink.new_keyset_handle(aead.aead_key_templates.AES128_GCM)
    encrypted_keyset = tink.json_proto_keyset_format.serialize_encrypted(
        keyset_handle,
        keyset_encryption_aead,
        keyset_encryption_associated_data,
    )

    reader = tink.JsonKeysetReader(encrypted_keyset)
    parsed_keyset_handle = tink.read_keyset_handle_with_associated_data(
        reader, keyset_encryption_aead, keyset_encryption_associated_data
    )

    # check that keyset_handle and parse_handle are the same.
    plaintext = b'plaintext'
    associated_data = b'associated_data'
    primitive1 = keyset_handle.primitive(aead.Aead)
    ciphertext = primitive1.encrypt(plaintext, associated_data)
    primitive2 = parsed_keyset_handle.primitive(aead.Aead)
    self.assertEqual(primitive2.decrypt(ciphertext, associated_data), plaintext)

  def test_write_with_associated_data_parse_encrypted(self):
    keyset_encryption_aead = tink.new_keyset_handle(
        aead.aead_key_templates.AES128_GCM
    ).primitive(aead.Aead)

    keyset_encryption_associated_data = b'keyset_encryption_associated_data'
    keyset_handle = tink.new_keyset_handle(aead.aead_key_templates.AES128_GCM)

    output_stream = io.StringIO()
    writer = tink.JsonKeysetWriter(output_stream)
    keyset_handle.write_with_associated_data(
        writer, keyset_encryption_aead, keyset_encryption_associated_data
    )
    encrypted_keyset = output_stream.getvalue()

    parsed_keyset_handle = tink.json_proto_keyset_format.parse_encrypted(
        encrypted_keyset,
        keyset_encryption_aead,
        keyset_encryption_associated_data,
    )

    # check that keyset_handle and parse_handle are the same.
    plaintext = b'plaintext'
    associated_data = b'associated_data'
    primitive1 = keyset_handle.primitive(aead.Aead)
    ciphertext = primitive1.encrypt(plaintext, associated_data)
    primitive2 = parsed_keyset_handle.primitive(aead.Aead)
    self.assertEqual(primitive2.decrypt(ciphertext, associated_data), plaintext)

  def test_serialize_encrypted_read_keyset_handle(self):
    keyset_encryption_aead = tink.new_keyset_handle(
        aead.aead_key_templates.AES128_GCM
    ).primitive(aead.Aead)

    # read_keyset_handle uses empty associated_data
    empty_keyset_encryption_associated_data = b''
    keyset_handle = tink.new_keyset_handle(aead.aead_key_templates.AES128_GCM)
    encrypted_keyset = tink.json_proto_keyset_format.serialize_encrypted(
        keyset_handle,
        keyset_encryption_aead,
        empty_keyset_encryption_associated_data,
    )

    reader = tink.JsonKeysetReader(encrypted_keyset)
    parsed_keyset_handle = tink.read_keyset_handle(
        reader, keyset_encryption_aead
    )

    # check that keyset_handle and parse_handle are the same.
    plaintext = b'plaintext'
    associated_data = b'associated_data'
    primitive1 = keyset_handle.primitive(aead.Aead)
    ciphertext = primitive1.encrypt(plaintext, associated_data)
    primitive2 = parsed_keyset_handle.primitive(aead.Aead)
    self.assertEqual(primitive2.decrypt(ciphertext, associated_data), plaintext)

  def test_write_parse_encrypted(self):
    keyset_encryption_aead = tink.new_keyset_handle(
        aead.aead_key_templates.AES128_GCM
    ).primitive(aead.Aead)

    keyset_handle = tink.new_keyset_handle(aead.aead_key_templates.AES128_GCM)

    output_stream = io.StringIO()
    writer = tink.JsonKeysetWriter(output_stream)
    keyset_handle.write(writer, keyset_encryption_aead)
    encrypted_keyset = output_stream.getvalue()

    # keyset_handle.write uses empty associated_data
    empty_keyset_encryption_associated_data = b''
    parsed_keyset_handle = tink.json_proto_keyset_format.parse_encrypted(
        encrypted_keyset,
        keyset_encryption_aead,
        empty_keyset_encryption_associated_data,
    )

    # check that keyset_handle and parse_handle are the same.
    plaintext = b'plaintext'
    associated_data = b'associated_data'
    primitive1 = keyset_handle.primitive(aead.Aead)
    ciphertext = primitive1.encrypt(plaintext, associated_data)
    primitive2 = parsed_keyset_handle.primitive(aead.Aead)
    self.assertEqual(primitive2.decrypt(ciphertext, associated_data), plaintext)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tink.python.tink.aead_key_templates."""

from absl.testing import absltest
from absl.testing import parameterized

from tink.proto import aes_ctr_hmac_aead_pb2
from tink.proto import aes_eax_pb2
from tink.proto import aes_gcm_pb2
from tink.proto import aes_gcm_siv_pb2
from tink.proto import common_pb2
from tink.proto import kms_aead_pb2
from tink.proto import kms_envelope_pb2
from tink.proto import tink_pb2
from tink.proto import x_aes_gcm_pb2
from tink import aead


class AeadKeyTemplatesTest(parameterized.TestCase):

  def test_create_kms_aead_key_template(self):
    template = aead.aead_key_templates.create_kms_aead_key_template(
        key_uri='fake://kek/uri'
    )
    self.assertEqual(template.type_url,
                     'type.googleapis.com/google.crypto.tink.KmsAeadKey')
    self.assertEqual(template.output_prefix_type, tink_pb2.RAW)
    key_format = kms_aead_pb2.KmsAeadKeyFormat.FromString(template.value)
    self.assertEqual(key_format.key_uri, 'fake://kek/uri')

  def test_create_kms_envelope_aead_key_template(self):
    template = aead.aead_key_templates.create_kms_envelope_aead_key_template(
        kek_uri='fake://kek/uri',
        dek_template=aead.aead_key_templates.AES128_GCM)
    self.assertEqual(
        template.type_url,
        'type.googleapis.com/google.crypto.tink.KmsEnvelopeAeadKey')
    self.assertEqual(template.output_prefix_type, tink_pb2.RAW)
    key_format = kms_envelope_pb2.KmsEnvelopeAeadKeyFormat.FromString(
        template.value)
    self.assertEqual(key_format.kek_uri, 'fake://kek/uri')
    self.assertEqual(key_format.dek_template.type_url,
                     aead.aead_key_templates.AES128_GCM.type_url)

  def test_create_aes_eax_key_template(self):
    # Intentionally using 'weird' or invalid values for parameters,
    # to test that the function correctly puts them in the resulting template.
    template = None
    with self.assertWarns(DeprecationWarning):
      template = aead.aead_key_templates.create_aes_eax_key_template(
          key_size=42, iv_size=72)
    self.assertEqual('type.googleapis.com/google.crypto.tink.AesEaxKey',
                     template.type_url)
    self.assertEqual(tink_pb2.TINK, template.output_prefix_type)
    key_format = aes_eax_pb2.AesEaxKeyFormat.FromString(template.value)
    self.assertEqual(42, key_format.key_size)
    self.assertEqual(72, key_format.params.iv_size)

  def test_create_aes_gcm_key_template(self):
    # Intentionally using 'weird' or invalid values for parameters,
    # to test that the function correctly puts them in the resulting template.
    template = None
    with self.assertWarns(DeprecationWarning):
      template = aead.aead_key_templates.create_aes_gcm_key_template(
          key_size=42)
    self.assertEqual('type.googleapis.com/google.crypto.tink.AesGcmKey',
                     template.type_url)
    self.assertEqual(tink_pb2.TINK, template.output_prefix_type)
    key_format = aes_gcm_pb2.AesGcmKeyFormat.FromString(template.value)
    self.assertEqual(42, key_format.key_size)

  def test_create_aes_gcm_siv_key_template(self):
    template = None
    with self.assertWarns(DeprecationWarning):
      template = aead.aead_key_templates.create_aes_gcm_siv_key_template(
          key_size=42)
    self.assertEqual('type.googleapis.com/google.crypto.tink.AesGcmSivKey',
                     template.type_url)
    self.assertEqual(tink_pb2.TINK, template.output_prefix_type)
    key_format = aes_gcm_siv_pb2.AesGcmSivKeyFormat.FromString(template.value)
    self.assertEqual(42, key_format.key_size)

  def test_create_aes_ctr_hmac_aead_key_template(self):
    # Intentionally using 'weird' or invalid values for parameters,
    # to test that the function correctly puts them in the resulting template.
    template = None
    with self.assertWarns(DeprecationWarning):
      template = aead.aead_key_templates.create_aes_ctr_hmac_aead_key_template(
          aes_key_size=34,
          iv_size=93,
          hmac_key_size=46,
          tag_size=99,
          hash_type=common_pb2.SHA1)
    self.assertEqual('type.googleapis.com/google.crypto.tink.AesCtrHmacAeadKey',
                     template.type_url)
    self.assertEqual(tink_pb2.TINK, template.output_prefix_type)
    key_format = aes_ctr_hmac_aead_pb2.AesCtrHmacAeadKeyFormat.FromString(
        template.value)
    self.assertEqual(93, key_format.aes_ctr_key_format.params.iv_size)
    self.assertEqual(34, key_format.aes_ctr_key_format.key_size)
    self.assertEqual(common_pb2.SHA1, key_format.hmac_key_format.params.hash)
    self.assertEqual(99, key_format.hmac_key_format.params.tag_size)
    self.assertEqual(46, key_format.hmac_key_format.key_size)

  def test_create_x_aes_gcm_key_template(self):

    for tc in [
        (aead.aead_key_templates.XAES_256_GCM_192_BIT_NONCE, 12, tink_pb2.TINK),
        (
            aead.aead_key_templates.XAES_256_GCM_192_BIT_NONCE_NO_PREFIX,
            12,
            tink_pb2.RAW,
        ),
        (
            aead.aead_key_templates.XAES_256_GCM_160_BIT_NONCE_NO_PREFIX,
            8,
            tink_pb2.RAW,
        ),
    ]:
      template = tc[0]
      self.assertEqual(
          'type.googleapis.com/google.crypto.tink.XAesGcmKey',
          template.type_url,
      )
      self.assertEqual(template.output_prefix_type, tc[2])
      key_format = x_aes_gcm_pb2.XAesGcmKeyFormat.FromString(template.value)
      self.assertEqual(tc[1], key_format.params.salt_size)

if __name__ == '__main__':
  absltest.main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: tink/proto/jwt_rsa_ssa_pss.proto
# Protobuf Python Version: 6.30.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    30,
    2,
    '',
    'tink/proto/jwt_rsa_ssa_pss.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n tink/proto/jwt_rsa_ssa_pss.proto\x12\x12google.crypto.tink\"\xe1\x01\n\x15JwtRsaSsaPssPublicKey\x12\x0f\n\x07version\x18\x01 \x01(\r\x12<\n\talgorithm\x18\x02 \x01(\x0e\x32).google.crypto.tink.JwtRsaSsaPssAlgorithm\x12\t\n\x01n\x18\x03 \x01(\x0c\x12\t\n\x01\x65\x18\x04 \x01(\x0c\x12G\n\ncustom_kid\x18\x05 \x01(\x0b\x32\x33.google.crypto.tink.JwtRsaSsaPssPublicKey.CustomKid\x1a\x1a\n\tCustomKid\x12\r\n\x05value\x18\x01 \x01(\t\"\xae\x01\n\x16JwtRsaSsaPssPrivateKey\x12\x0f\n\x07version\x18\x01 \x01(\r\x12=\n\npublic_key\x18\x02 \x01(\x0b\x32).google.crypto.tink.JwtRsaSsaPssPublicKey\x12\t\n\x01\x64\x18\x03 \x01(\x0c\x12\t\n\x01p\x18\x04 \x01(\x0c\x12\t\n\x01q\x18\x05 \x01(\x0c\x12\n\n\x02\x64p\x18\x06 \x01(\x0c\x12\n\n\x02\x64q\x18\x07 \x01(\x0c\x12\x0b\n\x03\x63rt\x18\x08 \x01(\x0c\"\x9d\x01\n\x15JwtRsaSsaPssKeyFormat\x12\x0f\n\x07version\x18\x01 \x01(\r\x12<\n\talgorithm\x18\x02 \x01(\x0e\x32).google.crypto.tink.JwtRsaSsaPssAlgorithm\x12\x1c\n\x14modulus_size_in_bits\x18\x03 \x01(\r\x12\x17\n\x0fpublic_exponent\x18\x04 \x01(\x0c*H\n\x15JwtRsaSsaPssAlgorithm\x12\x0e\n\nPS_UNKNOWN\x10\x00\x12\t\n\x05PS256\x10\x01\x12\t\n\x05PS384\x10\x02\x12\t\n\x05PS512\x10\x03\x42\x62\n\x1c\x63om.google.crypto.tink.protoP\x01Z@github.com/tink-crypto/tink-go/v2/proto/jwt_rsa_ssa_pss_go_protob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tink.proto.jwt_rsa_ssa_pss_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\034com.google.crypto.tink.protoP\001Z@github.com/tink-crypto/tink-go/v2/proto/jwt_rsa_ssa_pss_go_proto'
  _globals['_JWTRSASSAPSSALGORITHM']._serialized_start=621
  _globals['_JWTRSASSAPSSALGORITHM']._serialized_end=693
  _globals['_JWTRSASSAPSSPUBLICKEY']._serialized_start=57
  _globals['_JWTRSASSAPSSPUBLICKEY']._serialized_end=282
  _globals['_JWTRSASSAPSSPUBLICKEY_CUSTOMKID']._serialized_start=256
  _globals['_JWTRSASSAPSSPUBLICKEY_CUSTOMKID']._serialized_end=282
  _globals['_JWTRSASSAPSSPRIVATEKEY']._serialized_start=285
  _globals['_JWTRSASSAPSSPRIVATEKEY']._serialized_end=459
  _globals['_JWTRSASSAPSSKEYFORMAT']._serialized_start=462
  _globals['_JWTRSASSAPSSKEYFORMAT']._serialized_end=619
# @@protoc_insertion_point(module_scope)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for tink.python.tink.util._file_object_adapter."""

import io

from absl.testing import absltest
from absl.testing.absltest import mock

from tink.streaming_aead import _file_object_adapter


class FileObjectAdapterTest(absltest.TestCase):

  def test_basic_write(self):
    file_object = io.BytesIO()
    adapter = _file_object_adapter.FileObjectAdapter(file_object)

    self.assertEqual(9, adapter.write(b'something'))
    self.assertEqual(b'something', file_object.getvalue())
    adapter.close()

  def test_multiple_write(self):
    file_object = io.BytesIO()
    adapter = _file_object_adapter.FileObjectAdapter(file_object)

    self.assertEqual(9, adapter.write(b'something'))
    self.assertEqual(3, adapter.write(b'123'))
    self.assertEqual(3, adapter.write(b'456'))
    self.assertEqual(b'something123456', file_object.getvalue())

  def test_write_after_close(self):
    file_object = io.BytesIO()
    adapter = _file_object_adapter.FileObjectAdapter(file_object)

    adapter.close()
    with self.assertRaises(ValueError):
      adapter.write(b'something')

  def test_write_returns_none_written_is_zero(self):
    file_object = mock.Mock()
    file_object.write = mock.Mock(return_value=None)
    adapter = _file_object_adapter.FileObjectAdapter(file_object)

    self.assertEqual(0, adapter.write(b''))

  def test_write_returns_negative_number_raises_error(self):
    file_object = mock.Mock()
    file_object.write = mock.Mock(return_value=-1)
    adapter = _file_object_adapter.FileObjectAdapter(file_object)

    with self.assertRaises(ValueError):
      adapter.write(b'something')

  def test_write_returns_too_large_number_raises_error(self):
    data = b'something'
    file_object = mock.Mock()
    file_object.write = mock.Mock(return_value=len(data) + 1)
    adapter = _file_object_adapter.FileObjectAdapter(file_object)

    with self.assertRaises(ValueError):
      adapter.write(b'something')

  def test_write_raises_blocking_error_returns_characters_written_success(self):
    file_object = mock.Mock()
    file_object.write = mock.Mock(side_effect=io.BlockingIOError(None, None, 5))
    adapter = _file_object_adapter.FileObjectAdapter(file_object)

    self.assertEqual(5, adapter.write(b'something'))

  def test_write_raises_invalid_blocking_error_raises_error(self):
    file_object = mock.Mock()
    file_object.write = mock.Mock(
        side_effect=io.BlockingIOError(None, None, 42)
    )
    adapter = _file_object_adapter.FileObjectAdapter(file_object)

    with self.assertRaises(ValueError):
      adapter.write(b'something')

  def test_partial_write(self):
    file_object = mock.Mock()
    file_object.write = mock.Mock(wraps=lambda data: len(data) - 1)
    adapter = _file_object_adapter.FileObjectAdapter(file_object)

    self.assertEqual(8, adapter.write(b'something'))

  def test_basic_read(self):
    file_object = io.BytesIO(b'something')
    adapter = _file_object_adapter.FileObjectAdapter(file_object)

    self.assertEqual(adapter.read(9), b'something')

  def test_multiple_read(self):
    file_object = io.BytesIO(b'something')
    adapter = _file_object_adapter.FileObjectAdapter(file_object)

    self.assertEqual(adapter.read(3), b'som')
    self.assertEqual(adapter.read(3), b'eth')
    self.assertEqual(adapter.read(3), b'ing')

  def test_read_returns_none(self):
    file_object = mock.Mock()
    file_object.read = mock.Mock(return_value=None)
    adapter = _file_object_adapter.FileObjectAdapter(file_object)

    self.assertEqual(adapter.read(10), b'')

  def test_read_eof(self):
    file_object = mock.Mock()
    file_object.read = mock.Mock(return_value=b'')
    adapter = _file_object_adapter.FileObjectAdapter(file_object)

    with self.assertRaises(EOFError):
      adapter.read(10)

  def test_read_size_0(self):
    file_object = io.BytesIO(b'something')
    adapter = _file_object_adapter.FileObjectAdapter(file_object)

    self.assertEqual(adapter.read(0), b'')

  def test_read_negative_size_fails(self):
    file_object = io.BytesIO(b'something')
    adapter = _file_object_adapter.FileObjectAdapter(file_object)
    with self.assertRaises(ValueError):
      adapter.read(-1)

  def test_read_raises_blocking_error(self):
    file_object = mock.Mock()
    file_object.read = mock.Mock(side_effect=io.BlockingIOError(None, None))
    adapter = _file_object_adapter.FileObjectAdapter(file_object)

    self.assertEqual(adapter.read(10), b'')


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pre-generated KeyTemplate for HybridEncryption.

One can use these templates to generate a new tink_pb2.Keyset with
tink_pb2.KeysetHandle. To generate a new keyset that contains a single
tink_pb2.HmacKey, one can do:
handle = keyset_handle.KeysetHandle(mac_key_templates.HMAC_SHA256_128BITTAG).
"""

import warnings

from tink.proto import common_pb2
from tink.proto import ecies_aead_hkdf_pb2
from tink.proto import hpke_pb2
from tink.proto import tink_pb2
from tink import aead


def _create_ecies_aead_hkdf_key_template(
    curve_type: common_pb2.EllipticCurveType,
    ec_point_format: common_pb2.EcPointFormat, hash_type: common_pb2.HashType,
    dem_key_template: tink_pb2.KeyTemplate) -> tink_pb2.KeyTemplate:
  """Creates an ECIES-AEAD-HKDF KeyTemplate, and fills in its values."""
  key_format = ecies_aead_hkdf_pb2.EciesAeadHkdfKeyFormat()
  key_format.params.kem_params.curve_type = curve_type
  key_format.params.kem_params.hkdf_hash_type = hash_type
  key_format.params.dem_params.aead_dem.CopyFrom(dem_key_template)
  key_format.params.ec_point_format = ec_point_format

  key_template = tink_pb2.KeyTemplate()
  key_template.type_url = (
      'type.googleapis.com/google.crypto.tink.EciesAeadHkdfPrivateKey')
  key_template.value = key_format.SerializeToString()
  key_template.output_prefix_type = tink_pb2.TINK
  return key_template


def _create_hpke_key_template(
    hpke_kem: hpke_pb2.HpkeKem, hpke_kdf: hpke_pb2.HpkeKdf,
    hpke_aead: hpke_pb2.HpkeAead,
    output_prefix_type: tink_pb2.OutputPrefixType) -> tink_pb2.KeyTemplate:
  """Creates an HPKE KeyTemplate, and fills in its values."""
  key_format = hpke_pb2.HpkeKeyFormat()
  key_format.params.kem = hpke_kem
  key_format.params.kdf = hpke_kdf
  key_format.params.aead = hpke_aead

  key_template = tink_pb2.KeyTemplate(
      type_url='type.googleapis.com/google.crypto.tink.HpkePrivateKey',
      value=key_format.SerializeToString(),
      output_prefix_type=output_prefix_type,
  )
  return key_template


ECIES_P256_HKDF_HMAC_SHA256_AES128_GCM = _create_ecies_aead_hkdf_key_template(
    curve_type=common_pb2.NIST_P256,
    ec_point_format=common_pb2.UNCOMPRESSED,
    hash_type=common_pb2.SHA256,
    dem_key_template=aead.aead_key_templates.AES128_GCM)

ECIES_P256_COMPRESSED_HKDF_HMAC_SHA256_AES128_GCM = _create_ecies_aead_hkdf_key_template(
    curve_type=common_pb2.NIST_P256,
    ec_point_format=common_pb2.COMPRESSED,
    hash_type=common_pb2.SHA256,
    dem_key_template=aead.aead_key_templates.AES128_GCM)

ECIES_P256_HKDF_HMAC_SHA256_AES128_CTR_HMAC_SHA256 = (
    _create_ecies_aead_hkdf_key_template(
        curve_type=common_pb2.NIST_P256,
        ec_point_format=common_pb2.UNCOMPRESSED,
        hash_type=common_pb2.SHA256,
        dem_key_template=aead.aead_key_templates.AES128_CTR_HMAC_SHA256))

ECIES_P256_COMPRESSED_HKDF_HMAC_SHA256_AES128_CTR_HMAC_SHA256 = (
    _create_ecies_aead_hkdf_key_template(
        curve_type=common_pb2.NIST_P256,
        ec_point_format=common_pb2.COMPRESSED,
        hash_type=common_pb2.SHA256,
        dem_key_template=aead.aead_key_templates.AES128_CTR_HMAC_SHA256))

DHKEM_X25519_HKDF_SHA256_HKDF_SHA256_AES_128_GCM = (
    _create_hpke_key_template(
        hpke_kem=hpke_pb2.DHKEM_X25519_HKDF_SHA256,
        hpke_kdf=hpke_pb2.HKDF_SHA256,
        hpke_aead=hpke_pb2.AES_128_GCM,
        output_prefix_type=tink_pb2.TINK))

DHKEM_X25519_HKDF_SHA256_HKDF_SHA256_AES_128_GCM_RAW = (
    _create_hpke_key_template(
        hpke_kem=hpke_pb2.DHKEM_X25519_HKDF_SHA256,
        hpke_kdf=hpke_pb2.HKDF_SHA256,
        hpke_aead=hpke_pb2.AES_128_GCM,
        output_prefix_type=tink_pb2.RAW))

DHKEM_X25519_HKDF_SHA256_HKDF_SHA256_AES_256_GCM = (
    _create_hpke_key_template(
        hpke_kem=hpke_pb2.DHKEM_X25519_HKDF_SHA256,
        hpke_kdf=hpke_pb2.HKDF_SHA256,
        hpke_aead=hpke_pb2.AES_256_GCM,
        output_prefix_type=tink_pb2.TINK))

DHKEM_X25519_HKDF_SHA256_HKDF_SHA256_AES_256_GCM_RAW = (
    _create_hpke_key_template(
        hpke_kem=hpke_pb2.DHKEM_X25519_HKDF_SHA256,
        hpke_kdf=hpke_pb2.HKDF_SHA256,
        hpke_aead=hpke_pb2.AES_256_GCM,
        output_prefix_type=tink_pb2.RAW))

DHKEM_X25519_HKDF_SHA256_HKDF_SHA256_CHACHA20_POLY1305 = (
    _create_hpke_key_template(
        hpke_kem=hpke_pb2.DHKEM_X25519_HKDF_SHA256,
        hpke_kdf=hpke_pb2.HKDF_SHA256,
        hpke_aead=hpke_pb2.CHACHA20_POLY1305,
        output_prefix_type=tink_pb2.TINK))

DHKEM_X25519_HKDF_SHA256_HKDF_SHA256_CHACHA20_POLY1305_RAW = (
    _create_hpke_key_template(
        hpke_kem=hpke_pb2.DHKEM_X25519_HKDF_SHA256,
        hpke_kdf=hpke_pb2.HKDF_SHA256,
        hpke_aead=hpke_pb2.CHACHA20_POLY1305,
        output_prefix_type=tink_pb2.RAW))


# Deprecated. Use the predefined constant templates above instead.
def create_ecies_aead_hkdf_key_template(
    curve_type: common_pb2.EllipticCurveType,
    ec_point_format: common_pb2.EcPointFormat, hash_type: common_pb2.HashType,
    dem_key_template: tink_pb2.KeyTemplate) -> tink_pb2.KeyTemplate:
  warnings.warn(
      'The "create_ecies_aead_hkdf_key_template" function is deprecated.',
      DeprecationWarning, 2)
  return _create_ecies_aead_hkdf_key_template(curve_type, ec_point_format,
                                              hash_type, dem_key_template)



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Integration Tests for Tink GCP KMS integration."""

import base64
import os

from absl.testing import absltest

import tink
from tink import _kms_clients
from tink import aead
from tink import secret_key_access
from tink.integration import gcpkms
from tink.testing import helper

CREDENTIAL_PATH = os.path.join(
    helper.tink_py_testdata_path(), 'gcp/credential.json'
)
BAD_CREDENTIAL_PATH = os.path.join(
    helper.tink_py_testdata_path(), 'gcp/credential_bad.json'
)

KEY_URI = 'gcp-kms://projects/tink-test-infrastructure/locations/global/keyRings/unit-and-integration-testing/cryptoKeys/aead-key'
LOCAL_KEY_URI = 'gcp-kms://projects/tink-test-infrastructure/locations/europe-west1/keyRings/unit-and-integration-test/cryptoKeys/aead-key'
KEY2_URI = 'gcp-kms://projects/tink-test-infrastructure/locations/global/keyRings/unit-and-integration-testing/cryptoKeys/aead2-key'

if 'TEST_SRCDIR' in os.environ:
  # Set root certificates for gRPC in Bazel Test which are needed on MacOS
  os.environ['GRPC_DEFAULT_SSL_ROOTS_FILE_PATH'] = os.path.join(
      os.environ['TEST_SRCDIR'], 'google_root_pem/file/downloaded')


def setUpModule():
  aead.register()


class GcpKmsIntegrationTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    # Make sure default credentials are not set.
    os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = ''

  def tearDown(self):
    super().tearDown()
    _kms_clients.reset_kms_clients()

  def test_aead_from_keyset_handle_for_key_uri_works(self):
    # Register client not bound to a key URI.
    gcpkms.GcpKmsClient.register_client('', CREDENTIAL_PATH)

    handle = tink.new_keyset_handle(
        aead.aead_key_templates.create_kms_aead_key_template(KEY_URI)
    )
    gcp_aead = handle.primitive(aead.Aead)

    ciphertext = gcp_aead.encrypt(b'plaintext', b'associated_data')
    self.assertEqual(
        b'plaintext', gcp_aead.decrypt(ciphertext, b'associated_data')
    )
    with self.assertRaises(tink.TinkError):
      gcp_aead.decrypt(ciphertext, b'invalid')

  def test_envelope_aead_from_keyset_handle(self):
    # Register client not bound to a key URI.
    gcpkms.GcpKmsClient.register_client('', CREDENTIAL_PATH)

    handle = tink.new_keyset_handle(
        aead.aead_key_templates.create_kms_envelope_aead_key_template(
            KEY_URI, aead.aead_key_templates.AES128_GCM_SIV
        )
    )
    envelope_aead = handle.primitive(aead.Aead)

    ciphertext = envelope_aead.encrypt(b'plaintext', b'associated_data')
    self.assertEqual(
        b'plaintext', envelope_aead.decrypt(ciphertext, b'associated_data')
    )
    with self.assertRaises(tink.TinkError):
      envelope_aead.decrypt(ciphertext, b'invalid')

  def test_get_aead_is_compatible_with_kms_aead_key(self):
    # Get aead directly from GCP KMS client.
    kms_aead = gcpkms.GcpKmsClient('', CREDENTIAL_PATH).get_aead(KEY_URI)

    # Get aead by registering a client, and create a keyset with a KmsAeadKey.
    gcpkms.GcpKmsClient.register_client('', CREDENTIAL_PATH)
    handle = tink.new_keyset_handle(
        aead.aead_key_templates.create_kms_aead_key_template(KEY_URI)
    )
    aead_from_kms_aead_key = handle.primitive(aead.Aead)

    # Verify that they are compatible.
    ciphertext = kms_aead.encrypt(b'plaintext', b'associated_data')
    self.assertEqual(
        b'plaintext',
        aead_from_kms_aead_key.decrypt(ciphertext, b'associated_data'),
    )

  def test_kms_envelope_aead_is_compatible(self):
    # Get aead directly from GCP KMS client, and create an EnvelopeAead.
    kms_aead = gcpkms.GcpKmsClient('', CREDENTIAL_PATH).get_aead(KEY_URI)
    envelope_aead = aead.KmsEnvelopeAead(
        aead.aead_key_templates.AES128_GCM_SIV, kms_aead
    )

    # Get envelope aead by registering a client, and create a keyset with a
    # KmsEnvelopeAeadKey.
    gcpkms.GcpKmsClient.register_client('', CREDENTIAL_PATH)

    handle = tink.new_keyset_handle(
        aead.aead_key_templates.create_kms_envelope_aead_key_template(
            KEY_URI, aead.aead_key_templates.AES128_GCM_SIV
        )
    )
    aead_from_kms_envelope_aead_key = handle.primitive(aead.Aead)

    # Verify that they are compatible.
    ciphertext = envelope_aead.encrypt(b'plaintext', b'associated_data')
    self.assertEqual(
        b'plaintext',
        aead_from_kms_envelope_aead_key.decrypt(ciphertext, b'associated_data'),
    )

  def test_aead_from_keyset_handle_for_key2_uri(self):
    # Register client not bound to a key URI.
    gcpkms.GcpKmsClient.register_client('', CREDENTIAL_PATH)

    handle = tink.new_keyset_handle(
        aead.aead_key_templates.create_kms_aead_key_template(KEY2_URI)
    )
    gcp_aead = handle.primitive(aead.Aead)

    ciphertext = gcp_aead.encrypt(b'plaintext', b'associated_data')
    self.assertEqual(
        b'plaintext', gcp_aead.decrypt(ciphertext, b'associated_data')
    )

  def test_aead_from_keyset_handle_with_invalid_key_uri_fails(self):
    # Register client not bound to a key URI.
    gcpkms.GcpKmsClient.register_client('', CREDENTIAL_PATH)

    handle = tink.new_keyset_handle(
        aead.aead_key_templates.create_kms_aead_key_template(
            'aws-kms://arn:aws:kms:us-west-2:acc:other/key3'
        )
    )
    with self.assertRaises(tink.TinkError):
      gcp_aead = handle.primitive(aead.Aead)
      gcp_aead.encrypt(b'plaintext', b'associated_data')

  def test_decrypt_ciphertext_encrypted_in_bigquery_using_a_wrapped_keyset(
      self,
  ):
    # This wrapped keyset was generated in BigQuery using this command:
    # DECLARE kms_key_uri STRING;
    # SET kms_key_uri =
    # 'gcp-kms://projects/tink-test-infrastructure/locations/us/keyRings/big-query-test-key/cryptoKeys/aead-key';
    # SELECT KEYS.NEW_WRAPPED_KEYSET(kms_key_uri, 'AEAD_AES_GCM_256')
    wrapped_keyset = base64.urlsafe_b64decode(
        'CiQAv82D2I7RT2gQRd/01m+Md8WAmOyehVog50vs5uPq2B+R36YSlQEAba+J9rC0gfmX9F'
        'Ss8PsWIpCVbIvPiflsaHRxq5GQjknVgYuJLIMDXlGhQBa3NrfJSmj1T/KDHQ3EzCcPAXtO'
        'AbAExZr/7jsgiCzo/YQINyPb2rGkW4ofo/BVyvhZ/Pk40iuPHv8Q/PXVrNsq3Y2vkkpsyb'
        '3QUhJZseURGjjeQnZde6i3EmvDejXhOZJ3XdQUjwgorA=='
    )

    # This ciphertext was generated in BigQuery using this command:
    # DECLARE kms_key_uri STRING;
    # DECLARE wrapped_key BYTES;
    # SET kms_key_uri =
    # 'gcp-kms://projects/tink-test-infrastructure/locations/us/keyRings/big-query-test-key/cryptoKeys/aead-key';
    # SET wrapped_key =
    # FROM_BASE64('CiQAv82D2I7RT2gQRd/01m+Md8WAmOyehVog50vs5uPq2B+R36YSlQEAba+J9rC0gfmX9FSs8PsWIpCVbIvPiflsaHRxq5GQjknVgYuJLIMDXlGhQBa3NrfJSmj1T/KDHQ3EzCcPAXtOAbAExZr/7jsgiCzo/YQINyPb2rGkW4ofo/BVyvhZ/Pk40iuPHv8Q/PXVrNsq3Y2vkkpsyb3QUhJZseURGjjeQnZde6i3EmvDejXhOZJ3XdQUjwgorA==');
    # SELECT AEAD.ENCRYPT(KEYS.KEYSET_CHAIN(kms_key_uri, wrapped_key),
    #     'elephant', 'animal') AS encrypted_animal;
    ciphertext = base64.urlsafe_b64decode(
        'AcpCNBevmQnr9momhlKKEyKDOCj5bMfizqC22N/hLZd58LFpC+r99C0='
    )

    key_uri = 'gcp-kms://projects/tink-test-infrastructure/locations/us/keyRings/big-query-test-key/cryptoKeys/aead-key'
    gcp_aead = gcpkms.GcpKmsClient('', CREDENTIAL_PATH).get_aead(key_uri)

    unwrapped_keyset = gcp_aead.decrypt(wrapped_keyset, b'')
    keyset_handle = tink.proto_keyset_format.parse(
        unwrapped_keyset, secret_key_access.TOKEN
    )
    primitive = keyset_handle.primitive(aead.Aead)
    decrypted = primitive.decrypt(ciphertext, b'animal')
    self.assertEqual(decrypted, b'elephant')

  def test_registration_client_bound_to_uri_works(self):
    # Register client bound to KEY_URI.
    gcpkms.GcpKmsClient.register_client(KEY_URI, CREDENTIAL_PATH)

    # Create a keyset handle for KEY_URI and use it. This works.
    handle = tink.new_keyset_handle(
        aead.aead_key_templates.create_kms_aead_key_template(KEY_URI)
    )
    gcp_aead = handle.primitive(aead.Aead)
    ciphertext = gcp_aead.encrypt(b'plaintext', b'associated_data')
    self.assertEqual(
        b'plaintext', gcp_aead.decrypt(ciphertext, b'associated_data')
    )

    # But it fails for LOCAL_KEY_URI, since the URI is different.
    with self.assertRaises(tink.TinkError):
      handle2 = tink.new_keyset_handle(
          aead.aead_key_templates.create_kms_aead_key_template(LOCAL_KEY_URI)
      )
      gcp_aead = handle2.primitive(aead.Aead)
      gcp_aead.encrypt(b'plaintext', b'associated_data')

  def test_registration_client_with_default_credentials_works(self):
    # Set default credentials, see
    # https://cloud.google.com/docs/authentication/application-default-credentials
    os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = CREDENTIAL_PATH

    # register_client with credentials_path=None will use the default
    # credentials.
    gcpkms.GcpKmsClient.register_client(key_uri=KEY2_URI, credentials_path=None)

    handle = tink.new_keyset_handle(
        aead.aead_key_templates.create_kms_aead_key_template(KEY2_URI)
    )
    gcp_aead = handle.primitive(aead.Aead)
    ciphertext = gcp_aead.encrypt(b'plaintext', b'associated_data')
    self.assertEqual(
        b'plaintext', gcp_aead.decrypt(ciphertext, b'associated_data')
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A client for Hashicorp Vault."""

import base64
import re
from typing import Tuple
import urllib.parse
import hvac
import tink
from tink import aead

# Matches strings like {mount}/keys/{key_name}.
_PATH_MATCHER = re.compile(
    r"""
    # starts with 0 or one '/'
    [/]?
    # the capture group for {mount}
    (
      # a sequence of one or more characters that aren't '/'
      [^/]+
      # zero or more strings that start with '/' followed by a sequence of one
      # or more characters that aren't '/'
      (?:/[^/]+)*
    )
    /keys/
    # the capture group for {key_name}
    (
      # a sequence of one or more characters that aren't '/'
      [^/]+
    )
    $""",
    re.VERBOSE,
)


def _get_endpoint_paths(key_path: str) -> Tuple[str, str]:
  """Extracts mount point and key name from the given key_path.

  The key_path is expected of the form "{mount}/keys/{key_name}". For example,
  "transit/keys/key-foo" will be transformed to "transit" and "key-foo", and
  "teams/billing/service/cipher/keys/key-bar" will be transformed into
  "teams/billing/service/cipher" and "key-bar".

  Args:
    key_path: Key path of the form "/{mount}/keys/{key_name}".

  Returns:
    Vault transit encrypt/decrypt mount point and transit key name.
  """
  escaped_path = urllib.parse.quote(key_path)
  # Make sure that we have a path of the form: /{mount}/keys/{key_name}.
  mount_and_key_name = _PATH_MATCHER.fullmatch(escaped_path)
  if not mount_and_key_name:
    raise tink.TinkError('Invalid key_path')
  return mount_and_key_name.groups()


def new_aead(key_path: str, client: hvac.Client) -> aead.Aead:
  return _HcVaultKmsAead(client, key_path)


class _HcVaultKmsAead(aead.Aead):
  """Implements the Aead interface for Hashicorp Vault."""

  def __init__(self, client: hvac.Client, key_path: str) -> None:
    self.client = client
    mount_point, key_name = _get_endpoint_paths(key_path)
    self.key_name = key_name
    self.mount_point = mount_point

  def encrypt(self, plaintext: bytes, associated_data: bytes) -> bytes:
    try:
      if associated_data:
        raise NotImplementedError(
            'Tink HashiCorp Vault integration only allows empty associated'
            ' data. `associated_data` is not supported as of hvac==v2.1.0. See'
            ' https://github.com/hvac/hvac/issues/1107'
        )
      response = self.client.secrets.transit.encrypt_data(
          name=self.key_name,
          plaintext=base64.b64encode(plaintext).decode(),
          context='',  # Always empty. Tink assumes derived=false.
          mount_point=self.mount_point,
      )
      return response['data']['ciphertext'].encode()
    except Exception as e:
      raise tink.TinkError(e)

  def decrypt(self, ciphertext: bytes, associated_data: bytes) -> bytes:
    try:
      if associated_data:
        raise NotImplementedError(
            'Tink HashiCorp Vault integration only allows empty associated'
            ' data. `associated_data` is not supported as of hvac==v2.1.0. See'
            ' https://github.com/hvac/hvac/issues/1107'
        )
      response = self.client.secrets.transit.decrypt_data(
          name=self.key_name,
          ciphertext=ciphertext.decode(),
          context='',  # Always empty. Tink assumes derived=false.
          mount_point=self.mount_point,
      )
      return base64.b64decode(response['data']['plaintext'])
    except Exception as e:
      raise tink.TinkError(e)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implements a variant of BytesIO that lets you read the value after close().

This class can be used when an interface that writes to a stream and closes it
in the end need to be transformed into a function that returns a value.

An example is the implementation of normal AEAD encryption interface using
the streaming AEAD encryption interface.
"""

import errno
import io
from typing import Optional


class BytesIOWithValueAfterClose(io.BytesIO):
  """A BytesIO that lets you read the written value after close()."""

  def __init__(self, initial_bytes=None):
    self._finalvalue = None
    if initial_bytes:
      super().__init__(initial_bytes)
    else:
      super().__init__()

  def close(self) -> None:
    if not self.closed:
      self._value_after_close = self.getvalue()
    super().close()

  def value_after_close(self) -> bytes:
    if not self.closed:
      raise ValueError('call to value_after_close before close()')
    return self._value_after_close


class SlowBytesIO(BytesIOWithValueAfterClose):
  """An BytesIO that raised BlockingIOError on some calls to read or write."""

  def __init__(self, data: bytes = b'', seekable: bool = False):
    super().__init__(data)
    self._seekable = seekable
    self._state = -1

  def read(self, size: Optional[int] = -1) -> bytes:
    if size and size > 0:
      self._state += 1
      if self._state > 10000000:
        raise AssertionError('too many read/write. Is there an infinite loop?')
      if self._state % 3 == 0:   # block on every third call.
        raise io.BlockingIOError(
            errno.EAGAIN,
            'read could not complete without blocking', 0)
      # read at most 5 bytes.
      return super().read(min(size, 5))
    return super().read(size)

  def write(self, b: bytes) -> int:
    self._state += 1
    if self._state > 10000000:
      raise AssertionError('too many read/write. Is there an infinite loop?')
    if self._state % 3 == 0:
      # On every third call: write up to 3 bytes, then block.
      bytes_written = super().write(b[:3])
      raise io.BlockingIOError(
          errno.EAGAIN,
          'write could not complete without blocking',
          bytes_written,
      )
    # write at most 5 bytes.
    return super().write(b[:5])

  def seek(self, pos: int, whence: int = 0) -> int:
    if self._seekable:
      return super().seek(pos, whence)
    raise io.UnsupportedOperation('seek')

  def seekable(self)-> bool:
    return self._seekable


class SlowReadableRawBytes(io.RawIOBase):
  """A readable io.RawIOBase stream that only sometimes returns data."""

  def __init__(self, data: bytes, seekable: bool = False):
    super().__init__()
    self._bytes_io = io.BytesIO(data)
    self._seekable = seekable
    self._state = -1

  def readinto(self, b: bytearray) -> Optional[int]:
    try:
      self._state += 1
      if self._state > 10000000:
        raise AssertionError('too many read. Is there an infinite loop?')
      if self._state % 3 == 0:   # return None on every third call.
        return None
      # read at most 5 bytes
      q = self._bytes_io.read(5)
      b[:len(q)] = q
      return len(q)
    except io.BlockingIOError as exc:
      raise ValueError('io.BytesIO should not raise BlockingIOError') from exc

  def readable(self):
    return True

  def seek(self, pos: int, whence: int = 0) -> int:
    if self._seekable:
      return self._bytes_io.seek(pos, whence)
    raise io.UnsupportedOperation('seek')

  def seekable(self)-> bool:
    return self._seekable


class BadWriteBytesIO(BytesIOWithValueAfterClose):
  """An BytesIO that returns an invalid value from write()."""

  def write(self, b: bytes) -> int:
    # On every third call: write and return an invalid written length.
    bytes_written = super().write(b)
    return bytes_written + 1



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for envelope encryption with KMS."""

import struct

from tink.proto import tink_pb2
from tink import core
from tink.aead import _aead

_SUPPORTED_DEK_KEY_TYPES = frozenset({
    'type.googleapis.com/google.crypto.tink.AesGcmKey',
    'type.googleapis.com/google.crypto.tink.XChaCha20Poly1305Key',
    'type.googleapis.com/google.crypto.tink.AesCtrHmacAeadKey',
    'type.googleapis.com/google.crypto.tink.AesEaxKey',
    'type.googleapis.com/google.crypto.tink.AesGcmSivKey',
})

_MAX_ENCRYPTED_DEK_LEN = 4096


def is_supported_dek_key_type(type_url: str) -> bool:
  return type_url in _SUPPORTED_DEK_KEY_TYPES


class KmsEnvelopeAead(_aead.Aead):
  """Implements envelope encryption.

  Envelope encryption generates a data encryption key (DEK) which is used
  to encrypt the payload. The DEK is then send to a KMS to be encrypted and
  the encrypted DEK is attached to the ciphertext. In order to decrypt the
  ciphertext, the DEK first has to be decrypted by the KMS, and then the DEK
  can be used to decrypt the ciphertext. For further information see
  https://cloud.google.com/kms/docs/envelope-encryption.

  DEK key template must be a KeyTemplate for any of these Tink AEAD key types
  (any other key template will be rejected):
  * AesGcmKey
  * XChaCha20Poly1305
  * AesCtrHmacAeadKey
  * AesEaxKey
  * AesGcmSivKey

  The ciphertext structure is as follows:
  * Length of the encrypted DEK: 4 bytes (big endian)
  * Encrypted DEK: variable length, specified by the previous 4 bytes
  * AEAD payload: variable length
  """

  # Defines in how many bytes the DEK length will be encoded.
  DEK_LEN_BYTES = 4

  def __init__(self, key_template: tink_pb2.KeyTemplate, remote: _aead.Aead):
    if not is_supported_dek_key_type(key_template.type_url):
      raise core.TinkError(
          'Unsupported DEK key type: %s' % key_template.type_url
      )
    # Create a dek to make sure that it works, so that KmsEnvelopeAead already
    # fails when it is created, and not just when it is used.
    # The C++ implementation does the same check, and we want this
    # implementation to be consistent with the C++ implementation.
    _ = core.Registry.new_key_data(key_template)

    self.key_template = key_template
    self.remote_aead = remote

  def encrypt(self, plaintext: bytes, associated_data: bytes) -> bytes:
    # Get new key from template
    dek = core.Registry.new_key_data(self.key_template)
    dek_aead = core.Registry.primitive(dek, _aead.Aead)

    # Encrypt plaintext
    ciphertext = dek_aead.encrypt(plaintext, associated_data)

    # Wrap DEK key values with remote
    encrypted_dek = self.remote_aead.encrypt(dek.value, b'')

    if len(encrypted_dek) > _MAX_ENCRYPTED_DEK_LEN:
      raise core.TinkError('length of encrypted DEK too large')

    # Construct ciphertext, DEK length encoded as big endian
    enc_dek_len = struct.pack('>I', len(encrypted_dek))
    return enc_dek_len + encrypted_dek + ciphertext

  def decrypt(self, ciphertext: bytes, associated_data: bytes) -> bytes:
    ct_len = len(ciphertext)

    if ct_len < self.DEK_LEN_BYTES:
      raise core.TinkError

    # Parse first 4 bytes as big endian unsigned int
    dek_len = struct.unpack('>I', ciphertext[0:self.DEK_LEN_BYTES])[0]
    if (
        dek_len < 0
        or dek_len > _MAX_ENCRYPTED_DEK_LEN
        or dek_len > (ct_len - self.DEK_LEN_BYTES)
    ):
      raise core.TinkError('length of encrypted DEK too large')

    # Decrypt DEK with remote AEAD
    encrypted_dek_bytes = ciphertext[self.DEK_LEN_BYTES:self.DEK_LEN_BYTES +
                                     dek_len]
    dek_bytes = self.remote_aead.decrypt(encrypted_dek_bytes, b'')

    # Get AEAD primitive based on DEK
    dek = tink_pb2.KeyData(
        type_url=self.key_template.type_url,
        value=dek_bytes,
        key_material_type=tink_pb2.KeyData.SYMMETRIC,
    )
    dek_aead = core.Registry.primitive(dek, _aead.Aead)

    # Extract ciphertext payload and decrypt
    ct_bytes = ciphertext[self.DEK_LEN_BYTES + dek_len:]

    return dek_aead.decrypt(ct_bytes, associated_data)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Python implementation of a KeysetBuilder."""

import random
from tink.proto import tink_pb2
import tink
from tink import secret_key_access

_MAX_INT32 = 4294967295  # 2^32-1


def _new_key_data(key_template: tink_pb2.KeyTemplate) -> tink_pb2.KeyData:
  return tink.core.Registry.new_key_data(key_template)


def _generate_unused_key_id(keyset: tink_pb2.Keyset) -> int:
  while True:
    key_id = random.randint(1, _MAX_INT32)
    if key_id not in {key.key_id for key in keyset.key}:
      return key_id


def raw_template(template: tink_pb2.KeyTemplate) -> tink_pb2.KeyTemplate:
  template_copy = tink_pb2.KeyTemplate()
  template_copy.CopyFrom(template)
  template_copy.output_prefix_type = tink_pb2.RAW
  return template_copy


def legacy_template(template: tink_pb2.KeyTemplate) -> tink_pb2.KeyTemplate:
  template_copy = tink_pb2.KeyTemplate()
  template_copy.CopyFrom(template)
  template_copy.output_prefix_type = tink_pb2.LEGACY
  return template_copy


class KeysetBuilder:
  """A KeysetBuilder provides convenience methods for managing Keysets.

  It provides methods for adding, disabling, enabling, or deleting keys.
  The validity of the keyset is checked when creating a keyset_handle.
  """

  def __init__(self, keyset_proto: tink_pb2.Keyset):
    self._keyset = keyset_proto

  def keyset_handle(self) -> tink.KeysetHandle:
    return tink.proto_keyset_format.parse(
        self._keyset.SerializeToString(), secret_key_access.TOKEN
    )

  def keyset(self) -> bytes:
    return self._keyset.SerializeToString()

  def public_keyset(self) -> bytes:
    public_handle = self.keyset_handle().public_keyset_handle()
    return tink.proto_keyset_format.serialize_without_secret(public_handle)

  def add_new_key(self, key_template: tink_pb2.KeyTemplate) -> int:
    """Generates a new key, adds it to the keyset, and returns its ID."""
    new_key = self._keyset.key.add()
    new_key.key_data.CopyFrom(_new_key_data(key_template))
    new_key.status = tink_pb2.ENABLED
    new_key_id = _generate_unused_key_id(self._keyset)
    new_key.key_id = new_key_id
    new_key.output_prefix_type = key_template.output_prefix_type
    return new_key_id

  def set_primary_key(self, key_id: int) -> None:
    """Sets a key as primary."""
    for key in self._keyset.key:
      if key.key_id == key_id:
        self._keyset.primary_key_id = key_id
        return
    raise tink.TinkError('key not found: %d' % key_id)

  def enable_key(self, key_id: int) -> None:
    """Enables a key."""
    for key in self._keyset.key:
      if key.key_id == key_id:
        key.status = tink_pb2.ENABLED
        return
    raise tink.TinkError('key not found: %d' % key_id)

  def disable_key(self, key_id: int) -> None:
    """Disables a key."""
    for key in self._keyset.key:
      if key.key_id == key_id:
        key.status = tink_pb2.DISABLED
        return
    raise tink.TinkError('key not found: %d' % key_id)

  def delete_key(self, key_id: int) -> None:
    """Deletes a key."""
    for key in self._keyset.key:
      if key.key_id == key_id:
        self._keyset.key.remove(key)
        return
    raise tink.TinkError('key not found: %d' % key_id)


def from_keyset(keyset: bytes) -> KeysetBuilder:
  """Return a KeysetBuilder for a Keyset copied from a KeysetHandle."""
  keyset_proto = tink_pb2.Keyset.FromString(keyset)
  return KeysetBuilder(keyset_proto)


def from_keyset_handle(keyset_handle: tink.KeysetHandle) -> KeysetBuilder:
  """Return a KeysetBuilder for a Keyset copied from a KeysetHandle."""
  serialized_keyset = tink.proto_keyset_format.serialize(
      keyset_handle, secret_key_access.TOKEN
  )
  return from_keyset(serialized_keyset)


def new_keyset_builder() -> KeysetBuilder:
  return KeysetBuilder(tink_pb2.Keyset())



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module defines basic exceptions in Tink."""

from tink.cc.pybind import tink_bindings

KNOWN_STATUS_NOT_OK_TYPES = (tink_bindings.PythonTinkException,)


def register_status_not_ok_type(status_not_ok_type):
  global KNOWN_STATUS_NOT_OK_TYPES
  if status_not_ok_type not in KNOWN_STATUS_NOT_OK_TYPES:
    assert issubclass(status_not_ok_type, Exception)
    KNOWN_STATUS_NOT_OK_TYPES += (status_not_ok_type,)


def use_tink_errors(func):
  """Transforms StatusNotOk errors into TinkErrors."""

  def wrapper(*args, **kwargs):
    try:
      return func(*args, **kwargs)
    except KNOWN_STATUS_NOT_OK_TYPES as e:
      raise TinkError(e)
  return wrapper


class TinkError(Exception):
  """Common exception in Tink."""



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This class implements helper functions for testing."""

import os
from typing import Mapping

from runfiles import Runfiles
from tink.proto import tink_pb2
from tink import aead
from tink import core
from tink import daead
from tink import hybrid
from tink import mac
from tink import prf
from tink import signature as pk_signature

_RELATIVE_TESTDATA_PATH = 'google3/third_party/tink/testdata'


def _get_test_src_dir() -> str:
  # Note: this implementation is rewritten by copybara
  return os.path.join(
      Runfiles.Create().Rlocation('tink_py/testdata'))


def tink_py_testdata_path() -> str:
  """Returns the path to the test data directory to be used for testing."""
  # List of pairs <Env. variable, Path>.
  testdata_paths = []
  if 'TINK_PYTHON_ROOT_PATH' in os.environ:
    testdata_paths.append(('TINK_PYTHON_ROOT_PATH',
                           os.path.join(os.environ['TINK_PYTHON_ROOT_PATH'],
                                        'testdata')))
  if 'TEST_SRCDIR' in os.environ:
    testdata_paths.append(('TEST_SRCDIR', _get_test_src_dir()))
  for env_variable, testdata_path in testdata_paths:
    # Return the first path that is encountered.
    if not os.path.exists(testdata_path):
      raise FileNotFoundError(f'Variable {env_variable} is set but has an ' +
                              f'invalid path {testdata_path}')
    return testdata_path
  raise ValueError('No path environment variable set among ' +
                   'TINK_PYTHON_ROOT_PATH, TEST_SRCDIR')


def fake_key(
    value: bytes = b'fakevalue',
    type_url: str = 'fakeurl',
    key_material_type: tink_pb2.KeyData.KeyMaterialType = tink_pb2.KeyData
    .SYMMETRIC,
    key_id: int = 1234,
    status: tink_pb2.KeyStatusType = tink_pb2.ENABLED,
    output_prefix_type: tink_pb2.OutputPrefixType = tink_pb2.TINK
) -> tink_pb2.Keyset.Key:
  """Returns a fake but valid key."""
  key = tink_pb2.Keyset.Key(
      key_id=key_id,
      status=status,
      output_prefix_type=output_prefix_type)
  key.key_data.type_url = type_url
  key.key_data.value = value
  key.key_data.key_material_type = key_material_type
  return key


class FakeMac(mac.Mac):
  """A fake MAC implementation."""

  def __init__(self, name: str = 'FakeMac'):
    self._name = name

  def compute_mac(self, data: bytes) -> bytes:
    return data + b'|' + self._name.encode()

  def verify_mac(self, mac_value: bytes, data: bytes) -> None:
    if mac_value != data + b'|' + self._name.encode():
      raise core.TinkError('invalid mac ' + mac_value.decode())


class FakeAead(aead.Aead):
  """A fake AEAD implementation."""

  def __init__(self, name: str = 'FakeAead'):
    self._name = name

  def encrypt(self, plaintext: bytes, associated_data: bytes) -> bytes:
    return plaintext + b'|' + associated_data + b'|' + self._name.encode()

  def decrypt(self, ciphertext: bytes, associated_data: bytes) -> bytes:
    data = ciphertext.split(b'|')
    if (len(data) < 3 or data[1] != associated_data or
        data[2] != self._name.encode()):
      raise core.TinkError('failed to decrypt ciphertext ' +
                           ciphertext.decode())
    return data[0]


class FakeDeterministicAead(daead.DeterministicAead):
  """A fake Deterministic AEAD implementation."""

  def __init__(self, name: str = 'FakeDeterministicAead'):
    self._name = name

  def encrypt_deterministically(self, plaintext: bytes,
                                associated_data: bytes) -> bytes:
    return plaintext + b'|' + associated_data + b'|' + self._name.encode()

  def decrypt_deterministically(self, ciphertext: bytes,
                                associated_data: bytes) -> bytes:
    data = ciphertext.split(b'|')
    if (len(data) < 3 or
        data[1] != associated_data or
        data[2] != self._name.encode()):
      raise core.TinkError('failed to decrypt ciphertext ' +
                           ciphertext.decode())
    return data[0]


class FakeHybridDecrypt(hybrid.HybridDecrypt):
  """A fake HybridEncrypt implementation."""

  def __init__(self, name: str = 'Hybrid'):
    self._name = name

  def decrypt(self, ciphertext: bytes, context_info: bytes) -> bytes:
    data = ciphertext.split(b'|')
    if (len(data) < 3 or
        data[1] != context_info or
        data[2] != self._name.encode()):
      raise core.TinkError('failed to decrypt ciphertext ' +
                           ciphertext.decode())
    return data[0]


class FakeHybridEncrypt(hybrid.HybridEncrypt):
  """A fake HybridEncrypt implementation."""

  def __init__(self, name: str = 'Hybrid'):
    self._name = name

  def encrypt(self, plaintext: bytes, context_info: bytes) -> bytes:
    return plaintext + b'|' + context_info + b'|' + self._name.encode()


class FakePublicKeySign(pk_signature.PublicKeySign):
  """A fake PublicKeySign implementation."""

  def __init__(self, name: str = 'FakePublicKeySign'):
    self._name = name

  def sign(self, data: bytes) -> bytes:
    return data + b'|' + self._name.encode()


class FakePublicKeyVerify(pk_signature.PublicKeyVerify):
  """A fake PublicKeyVerify implementation."""

  def __init__(self, name: str = 'FakePublicKeyVerify'):
    self._name = name

  def verify(self, signature: bytes, data: bytes):
    if signature != data + b'|' + self._name.encode():
      raise core.TinkError('invalid signature ' + signature.decode())


class FakePrf(prf.Prf):
  """A fake Prf implementation."""

  def __init__(self, name: str = 'FakePrf'):
    self._name = name

  def compute(self, input_data: bytes, output_length: int) -> bytes:
    if output_length > 32:
      raise core.TinkError('invalid output_length')
    output = (
        input_data + b'|' + self._name.encode() + b'|' +
        b''.join([b'*' for _ in range(output_length)]))
    return output[:output_length]


class FakePrfSet(prf.PrfSet):
  """A fake PrfSet implementation that contains exactly one Prf."""

  def __init__(self, name: str = 'FakePrf'):
    self._prf = FakePrf(name)

  def primary_id(self) -> int:
    return 0

  def all(self) -> Mapping[int, prf.Prf]:
    return {0: self._prf}

  def primary(self) -> prf.Prf:
    return self._prf



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tink.python.tink.proto_keyset_format."""

import io

from absl.testing import absltest

import tink
from tink import aead
from tink import core
from tink import hybrid
from tink import mac
from tink import secret_key_access
from tink import tink_config


def setUpModule():
  tink_config.register()


class InvalidKeyAccess(core.KeyAccess):
  pass


class TinkProtoKeysetFormatTest(absltest.TestCase):

  def test_serialize_parse(self):
    keyset_handle = tink.new_keyset_handle(aead.aead_key_templates.AES128_GCM)
    serialized_keyset = tink.proto_keyset_format.serialize(
        keyset_handle, secret_key_access.TOKEN
    )
    parsed_keyset_handle = tink.proto_keyset_format.parse(
        serialized_keyset, secret_key_access.TOKEN
    )

    # check that keyset_handle and parse_handle are the same.
    plaintext = b'plaintext'
    associated_data = b'associated_data'
    primitive1 = keyset_handle.primitive(aead.Aead)
    ciphertext = primitive1.encrypt(plaintext, associated_data)
    primitive2 = parsed_keyset_handle.primitive(aead.Aead)
    self.assertEqual(primitive2.decrypt(ciphertext, associated_data), plaintext)

    # check that serialize and parse fail without secret_key_access.TOKEN
    with self.assertRaises(core.TinkError):
      tink.proto_keyset_format.serialize(keyset_handle, InvalidKeyAccess())
    with self.assertRaises(core.TinkError):
      tink.proto_keyset_format.parse(serialized_keyset, InvalidKeyAccess())

  def test_serialize_parse_without_secret(self):
    private_handle = tink.new_keyset_handle(
        hybrid.hybrid_key_templates.ECIES_P256_HKDF_HMAC_SHA256_AES128_GCM
    )
    public_handle = private_handle.public_keyset_handle()

    # serialize and parse public_handle
    serialized = tink.proto_keyset_format.serialize_without_secret(
        public_handle
    )
    parsed = tink.proto_keyset_format.parse_without_secret(serialized)

    # check that parsed works with private_handle
    plaintext = b'plaintext'
    context_info = b'context info'
    hybrid_enc = parsed.primitive(hybrid.HybridEncrypt)
    ciphertext = hybrid_enc.encrypt(plaintext, context_info)
    hybrid_dec = private_handle.primitive(hybrid.HybridDecrypt)
    self.assertEqual(hybrid_dec.decrypt(ciphertext, context_info), plaintext)

  def test_serialize_parse_encrypted(self):
    keyset_encryption_aead = tink.new_keyset_handle(
        aead.aead_key_templates.AES128_GCM
    ).primitive(aead.Aead)

    keyset_encryption_associated_data = b'keyset_encryption_associated_data'
    keyset_handle = tink.new_keyset_handle(aead.aead_key_templates.AES128_GCM)
    encrypted_keyset = tink.proto_keyset_format.serialize_encrypted(
        keyset_handle, keyset_encryption_aead, keyset_encryption_associated_data
    )
    parsed_keyset_handle = tink.proto_keyset_format.parse_encrypted(
        encrypted_keyset,
        keyset_encryption_aead,
        keyset_encryption_associated_data,
    )

    # check that keyset_handle and parse_handle are the same.
    plaintext = b'plaintext'
    associated_data = b'associated_data'
    primitive1 = keyset_handle.primitive(aead.Aead)
    ciphertext = primitive1.encrypt(plaintext, associated_data)
    primitive2 = parsed_keyset_handle.primitive(aead.Aead)
    self.assertEqual(primitive2.decrypt(ciphertext, associated_data), plaintext)

    with self.assertRaises(core.TinkError):
      tink.proto_keyset_format.parse_encrypted(
          encrypted_keyset,
          keyset_encryption_aead,
          b'invalid_associated_data',
      )

  def test_serialize_encrypted_read_keyset_handle_with_associated_data(self):
    keyset_encryption_aead = tink.new_keyset_handle(
        aead.aead_key_templates.AES128_GCM
    ).primitive(aead.Aead)

    keyset_encryption_associated_data = b'keyset_encryption_associated_data'
    keyset_handle = tink.new_keyset_handle(aead.aead_key_templates.AES128_GCM)
    encrypted_keyset = tink.proto_keyset_format.serialize_encrypted(
        keyset_handle,
        keyset_encryption_aead,
        keyset_encryption_associated_data,
    )

    reader = tink.BinaryKeysetReader(encrypted_keyset)
    parsed_keyset_handle = tink.read_keyset_handle_with_associated_data(
        reader, keyset_encryption_aead, keyset_encryption_associated_data
    )

    # check that keyset_handle and parse_handle are the same.
    plaintext = b'plaintext'
    associated_data = b'associated_data'
    primitive1 = keyset_handle.primitive(aead.Aead)
    ciphertext = primitive1.encrypt(plaintext, associated_data)
    primitive2 = parsed_keyset_handle.primitive(aead.Aead)
    self.assertEqual(primitive2.decrypt(ciphertext, associated_data), plaintext)

  def test_write_with_associated_data_parse_encrypted(self):
    keyset_encryption_aead = tink.new_keyset_handle(
        aead.aead_key_templates.AES128_GCM
    ).primitive(aead.Aead)

    keyset_encryption_associated_data = b'keyset_encryption_associated_data'
    keyset_handle = tink.new_keyset_handle(aead.aead_key_templates.AES128_GCM)

    output_stream = io.BytesIO()
    writer = tink.BinaryKeysetWriter(output_stream)
    keyset_handle.write_with_associated_data(
        writer, keyset_encryption_aead, keyset_encryption_associated_data
    )
    encrypted_keyset = output_stream.getvalue()

    parsed_keyset_handle = tink.proto_keyset_format.parse_encrypted(
        encrypted_keyset,
        keyset_encryption_aead,
        keyset_encryption_associated_data,
    )

    # check that keyset_handle and parse_handle are the same.
    plaintext = b'plaintext'
    associated_data = b'associated_data'
    primitive1 = keyset_handle.primitive(aead.Aead)
    ciphertext = primitive1.encrypt(plaintext, associated_data)
    primitive2 = parsed_keyset_handle.primitive(aead.Aead)
    self.assertEqual(primitive2.decrypt(ciphertext, associated_data), plaintext)

  def test_serialize_encrypted_read_keyset_handle(self):
    keyset_encryption_aead = tink.new_keyset_handle(
        aead.aead_key_templates.AES128_GCM
    ).primitive(aead.Aead)

    # read_keyset_handle uses empty associated_data
    empty_keyset_encryption_associated_data = b''
    keyset_handle = tink.new_keyset_handle(aead.aead_key_templates.AES128_GCM)
    encrypted_keyset = tink.proto_keyset_format.serialize_encrypted(
        keyset_handle,
        keyset_encryption_aead,
        empty_keyset_encryption_associated_data,
    )

    reader = tink.BinaryKeysetReader(encrypted_keyset)
    parsed_keyset_handle = tink.read_keyset_handle(
        reader, keyset_encryption_aead
    )

    # check that keyset_handle and parse_handle are the same.
    plaintext = b'plaintext'
    associated_data = b'associated_data'
    primitive1 = keyset_handle.primitive(aead.Aead)
    ciphertext = primitive1.encrypt(plaintext, associated_data)
    primitive2 = parsed_keyset_handle.primitive(aead.Aead)
    self.assertEqual(primitive2.decrypt(ciphertext, associated_data), plaintext)

  def test_write_parse_encrypted(self):
    keyset_encryption_aead = tink.new_keyset_handle(
        aead.aead_key_templates.AES128_GCM
    ).primitive(aead.Aead)

    keyset_handle = tink.new_keyset_handle(aead.aead_key_templates.AES128_GCM)

    output_stream = io.BytesIO()
    writer = tink.BinaryKeysetWriter(output_stream)
    keyset_handle.write(writer, keyset_encryption_aead)
    encrypted_keyset = output_stream.getvalue()

    # keyset_handle.write uses empty associated_data
    empty_keyset_encryption_associated_data = b''
    parsed_keyset_handle = tink.proto_keyset_format.parse_encrypted(
        encrypted_keyset,
        keyset_encryption_aead,
        empty_keyset_encryption_associated_data,
    )

    # check that keyset_handle and parse_handle are the same.
    plaintext = b'plaintext'
    associated_data = b'associated_data'
    primitive1 = keyset_handle.primitive(aead.Aead)
    ciphertext = primitive1.encrypt(plaintext, associated_data)
    primitive2 = parsed_keyset_handle.primitive(aead.Aead)
    self.assertEqual(primitive2.decrypt(ciphertext, associated_data), plaintext)

  def test_parse_keyset_from_test_vector(self):
    # Same as in TinkProtoKeysetFormatTest.parseKeysetFromTestVector.
    # It contains one HMAC key.
    serialized_keyset = bytes.fromhex(
        '0895e59bcc0612680a5c0a2e747970652e676f6f676c65617069732e636f6d2f676f6f'
        '676c652e63727970746f2e74696e6b2e486d61634b657912281a20cca20f02278003b3'
        '513f5d01759ac1302f7d883f2f4a40025532ee1b11f9e5871204101008031801100118'
        '95e59bcc062001'
    )
    parsed_keyset_handle = tink.proto_keyset_format.parse(
        serialized_keyset, secret_key_access.TOKEN
    )
    primitive = parsed_keyset_handle.primitive(mac.Mac)
    valid_mac = bytes.fromhex('016986f2956092d259136923c6f4323557714ec499')
    data = b'data'
    primitive.verify_mac(valid_mac, data)

  def test_parse_encrypted_keyset_from_test_vector(self):
    # Same as in TinkProtoKeysetFormatTest.parseEncryptedKeysetFromTestVector.
    # An AEAD key, with which we encrypted the mac keyset below.
    serialized_keyset_encryption_keyset = bytes.fromhex(
        '08b891f5a20412580a4c0a30747970652e676f6f676c65617069732e636f6d2f676f6f'
        '676c652e63727970746f2e74696e6b2e4165734561784b65791216120208101a10e5d7'
        'd0cdd649e81e7952260689b2e1971801100118b891f5a2042001'
    )
    keyset_encryption_handle = tink.proto_keyset_format.parse(
        serialized_keyset_encryption_keyset, secret_key_access.TOKEN
    )
    keyset_encryption_aead = keyset_encryption_handle.primitive(aead.Aead)

    # A keyset that contains one HMAC key, encrypted with the above, using
    # associatedData
    encrypted_serialized_keyset = bytes.fromhex(
        '12950101445d48b8b5f591efaf73a46df9ebd7b6ac471cc0cf4f815a4f012fcaffc8f0'
        'b2b10b30c33194f0b291614bd8e1d2e80118e5d6226b6c41551e104ef8cd8ee20f1c14'
        'c1b87f6eed5fb04a91feafaacbf6f368519f36f97f7d08b24c8e71b5e620c4f69615ef'
        '0479391666e2fb32e46b416893fc4e564ba927b22ebff2a77bd3b5b8d5afa162cbd35c'
        '94c155cdfa13c8a9c964cde21a4208f5909ce901123a0a2e747970652e676f6f676c65'
        '617069732e636f6d2f676f6f676c652e63727970746f2e74696e6b2e486d61634b6579'
        '100118f5909ce9012001'
    )
    associated_data = bytes.fromhex('abcdef330012')

    handle = tink.proto_keyset_format.parse_encrypted(
        encrypted_serialized_keyset,
        keyset_encryption_aead,
        associated_data,
    )

    primitive = handle.primitive(mac.Mac)
    message = b''
    tag = bytes.fromhex('011d270875989dd6fbd5f54dbc9520bb41efd058d5')
    primitive.verify_mac(tag, message)

  def test_encrypted_keyset_overhead(self):
    keyset_encryption_aead = tink.new_keyset_handle(
        aead.aead_key_templates.AES128_GCM
    ).primitive(aead.Aead)

    keyset_handle = tink.new_keyset_handle(aead.aead_key_templates.AES128_GCM)

    serialized_keyset = tink.proto_keyset_format.serialize(
        keyset_handle, secret_key_access.TOKEN
    )
    raw_encrypted_keyset = keyset_encryption_aead.encrypt(
        serialized_keyset, b''
    )

    encrypted_keyset = tink.proto_keyset_format.serialize_encrypted(
        keyset_handle=keyset_handle,
        keyset_encryption_aead=keyset_encryption_aead,
        associated_data=b'',
    )
    # encrypted_keyset is a serialized protocol buffer that contains only
    # raw_encrypted_keyset in a field. So
    # it should only be slightly larger than raw_encrypted_keyset.
    # The overhead is currently just 2 bytes, but we choose 6 here to avoid
    # making the test brittle or flaky.
    self.assertLessEqual(len(encrypted_keyset), len(raw_encrypted_keyset) + 6)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Public Key Sign wrapper."""

from typing import Optional, Type

from tink.proto import tink_pb2
from tink import _monitoring
from tink import core
from tink.signature import _public_key_sign
from tink.signature import _public_key_verify


class _WrappedPublicKeySign(_public_key_sign.PublicKeySign):
  """Implements PublicKeySign for a set of PublicKeySign primitives."""

  def __init__(
      self,
      primitives_set: core.PrimitiveSet,
      monitor: Optional[_monitoring.KeyUsageMonitor] = None,
  ):
    self._primitive_set = primitives_set
    self._monitor = monitor

  def sign(self, data: bytes) -> bytes:
    """Computes the signature for data using the primary primitive.

    Args:
      data: The input data.

    Returns:
      The signature.
    """
    primary = self._primitive_set.primary()

    if not primary:
      raise core.TinkError('primary primitive not set')

    sign_data = data
    if primary.output_prefix_type == tink_pb2.LEGACY:
      sign_data = sign_data + b'\x00'

    result = primary.identifier + primary.primitive.sign(sign_data)

    if self._monitor:
      self._monitor.log(primary.key_id, len(sign_data))

    return result


class PublicKeySignWrapper(
    core.PrimitiveWrapper[_public_key_sign.PublicKeySign,
                          _public_key_sign.PublicKeySign]):
  """A PrimitiveWrapper for the PublicKeySign primitive.

  The returned primitive works with a keyset (rather than a single key). To sign
  a message, it uses the primary key in the keyset, and prepends to the
  signature a certain prefix associated with the primary key.
  """

  def wrap(self, primitives_set: core.PrimitiveSet) -> _WrappedPublicKeySign:
    return _WrappedPublicKeySign(primitives_set)

  def primitive_class(self) -> Type[_public_key_sign.PublicKeySign]:
    return _public_key_sign.PublicKeySign

  def input_primitive_class(self) -> Type[_public_key_sign.PublicKeySign]:
    return _public_key_sign.PublicKeySign

  def _wrap_with_monitoring_info(
      self,
      primitives_set: core.PrimitiveSet,
      monitoring_keyset_info: _monitoring.MonitoringKeySetInfo,
  ) -> _public_key_sign.PublicKeySign:
    key_usage_monitor = _monitoring.get_key_usage_monitor_or_none(
        _monitoring.MonitoringContext(
            primitive='public_key_sign',
            api_function='sign',
            keyset_info=monitoring_keyset_info,
        )
    )
    return _WrappedPublicKeySign(primitives_set, key_usage_monitor)


class _WrappedPublicKeyVerify(_public_key_verify.PublicKeyVerify):
  """Implements PublicKeyVerify for a set of PublicKeyVerify primitives."""

  def __init__(
      self,
      primitives_set: core.PrimitiveSet,
      monitor: Optional[_monitoring.KeyUsageMonitor] = None,
  ):
    self._primitive_set = primitives_set
    self._monitor = monitor

  def verify(self, signature: bytes, data: bytes):
    """Verifies that signature is a digital signature for data.

    Args:
      signature: The signature bytes to be checked.
      data: The data bytes to be checked.

    Raises:
      tink_error.TinkError if the verification fails.
    """
    if len(signature) <= core.crypto_format.NON_RAW_PREFIX_SIZE:
      if self._monitor:
        self._monitor.log_failure()
      # This also rejects raw signatures with size of 4 bytes or fewer.
      # We're not aware of any schemes that output signatures that small.
      raise core.TinkError('signature too short')

    key_id = signature[:core.crypto_format.NON_RAW_PREFIX_SIZE]
    raw_sig = signature[core.crypto_format.NON_RAW_PREFIX_SIZE:]

    for entry in self._primitive_set.primitive_from_identifier(key_id):
      try:
        if entry.output_prefix_type == tink_pb2.LEGACY:
          entry.primitive.verify(raw_sig, data + b'\x00')
        else:
          entry.primitive.verify(raw_sig, data)
        # Signature is valid, we can return
        if self._monitor:
          self._monitor.log(entry.key_id, len(signature))
        return
      except core.TinkError:
        pass

    # No matching key succeeded with verification, try all RAW keys
    for entry in self._primitive_set.raw_primitives():
      try:
        entry.primitive.verify(signature, data)
        # Signature is valid, we can return
        if self._monitor:
          self._monitor.log(entry.key_id, len(signature))
        return
      except core.TinkError:
        pass

    # nothing works.
    if self._monitor:
      self._monitor.log_failure()

    raise core.TinkError('invalid signature')


class PublicKeyVerifyWrapper(
    core.PrimitiveWrapper[_public_key_verify.PublicKeyVerify,
                          _public_key_verify.PublicKeyVerify]):
  """WrappedPublicKeyVerify is the PrimitiveWrapper for PublicKeyVerify.

  The returned primitive works with a keyset (rather than a single key). To sign
  a message, it uses the primary key in the keyset, and prepends to the
  signature a certain prefix associated with the primary key.

  The returned primitive works with a keyset (rather than a single key). To
  verify a signature, the primitive uses the prefix of the signature to
  efficiently select the right key in the set. If there is no key associated
  with the prefix or if the keys associated with the prefix do not work, the
  primitive tries all keys with tink_pb2.OutputPrefixType = tink_pb2.RAW.
  """

  def wrap(self, primitives_set: core.PrimitiveSet) -> _WrappedPublicKeyVerify:
    return _WrappedPublicKeyVerify(primitives_set)

  def primitive_class(self) -> Type[_public_key_verify.PublicKeyVerify]:
    return _public_key_verify.PublicKeyVerify

  def input_primitive_class(self) -> Type[_public_key_verify.PublicKeyVerify]:
    return _public_key_verify.PublicKeyVerify

  def _wrap_with_monitoring_info(
      self,
      primitives_set: core.PrimitiveSet,
      monitoring_keyset_info: _monitoring.MonitoringKeySetInfo,
  ) -> _public_key_verify.PublicKeyVerify:
    key_usage_monitor = _monitoring.get_key_usage_monitor_or_none(
        _monitoring.MonitoringContext(
            primitive='public_key_verify',
            api_function='verify',
            keyset_info=monitoring_keyset_info,
        )
    )
    return _WrappedPublicKeyVerify(primitives_set, key_usage_monitor)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: tink/proto/rsa_ssa_pss.proto
# Protobuf Python Version: 6.30.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    30,
    2,
    '',
    'tink/proto/rsa_ssa_pss.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tink.proto import common_pb2 as tink_dot_proto_dot_common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1ctink/proto/rsa_ssa_pss.proto\x12\x12google.crypto.tink\x1a\x17tink/proto/common.proto\"\x87\x01\n\x0fRsaSsaPssParams\x12.\n\x08sig_hash\x18\x01 \x01(\x0e\x32\x1c.google.crypto.tink.HashType\x12/\n\tmgf1_hash\x18\x02 \x01(\x0e\x32\x1c.google.crypto.tink.HashType\x12\x13\n\x0bsalt_length\x18\x03 \x01(\x05\"p\n\x12RsaSsaPssPublicKey\x12\x0f\n\x07version\x18\x01 \x01(\r\x12\x33\n\x06params\x18\x02 \x01(\x0b\x32#.google.crypto.tink.RsaSsaPssParams\x12\t\n\x01n\x18\x03 \x01(\x0c\x12\t\n\x01\x65\x18\x04 \x01(\x0c\"\xa8\x01\n\x13RsaSsaPssPrivateKey\x12\x0f\n\x07version\x18\x01 \x01(\r\x12:\n\npublic_key\x18\x02 \x01(\x0b\x32&.google.crypto.tink.RsaSsaPssPublicKey\x12\t\n\x01\x64\x18\x03 \x01(\x0c\x12\t\n\x01p\x18\x04 \x01(\x0c\x12\t\n\x01q\x18\x05 \x01(\x0c\x12\n\n\x02\x64p\x18\x06 \x01(\x0c\x12\n\n\x02\x64q\x18\x07 \x01(\x0c\x12\x0b\n\x03\x63rt\x18\x08 \x01(\x0c\"\x80\x01\n\x12RsaSsaPssKeyFormat\x12\x33\n\x06params\x18\x01 \x01(\x0b\x32#.google.crypto.tink.RsaSsaPssParams\x12\x1c\n\x14modulus_size_in_bits\x18\x02 \x01(\r\x12\x17\n\x0fpublic_exponent\x18\x03 \x01(\x0c\x42^\n\x1c\x63om.google.crypto.tink.protoP\x01Z<github.com/tink-crypto/tink-go/v2/proto/rsa_ssa_pss_go_protob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tink.proto.rsa_ssa_pss_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\034com.google.crypto.tink.protoP\001Z<github.com/tink-crypto/tink-go/v2/proto/rsa_ssa_pss_go_proto'
  _globals['_RSASSAPSSPARAMS']._serialized_start=78
  _globals['_RSASSAPSSPARAMS']._serialized_end=213
  _globals['_RSASSAPSSPUBLICKEY']._serialized_start=215
  _globals['_RSASSAPSSPUBLICKEY']._serialized_end=327
  _globals['_RSASSAPSSPRIVATEKEY']._serialized_start=330
  _globals['_RSASSAPSSPRIVATEKEY']._serialized_end=498
  _globals['_RSASSAPSSKEYFORMAT']._serialized_start=501
  _globals['_RSASSAPSSKEYFORMAT']._serialized_end=629
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: tink/proto/aes_ctr_hmac_aead.proto
# Protobuf Python Version: 6.30.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    30,
    2,
    '',
    'tink/proto/aes_ctr_hmac_aead.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tink.proto import aes_ctr_pb2 as tink_dot_proto_dot_aes__ctr__pb2
from tink.proto import hmac_pb2 as tink_dot_proto_dot_hmac__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\"tink/proto/aes_ctr_hmac_aead.proto\x12\x12google.crypto.tink\x1a\x18tink/proto/aes_ctr.proto\x1a\x15tink/proto/hmac.proto\"\x96\x01\n\x17\x41\x65sCtrHmacAeadKeyFormat\x12?\n\x12\x61\x65s_ctr_key_format\x18\x01 \x01(\x0b\x32#.google.crypto.tink.AesCtrKeyFormat\x12:\n\x0fhmac_key_format\x18\x02 \x01(\x0b\x32!.google.crypto.tink.HmacKeyFormat\"\x87\x01\n\x11\x41\x65sCtrHmacAeadKey\x12\x0f\n\x07version\x18\x01 \x01(\r\x12\x32\n\x0b\x61\x65s_ctr_key\x18\x02 \x01(\x0b\x32\x1d.google.crypto.tink.AesCtrKey\x12-\n\x08hmac_key\x18\x03 \x01(\x0b\x32\x1b.google.crypto.tink.HmacKeyBd\n\x1c\x63om.google.crypto.tink.protoP\x01ZBgithub.com/tink-crypto/tink-go/v2/proto/aes_ctr_hmac_aead_go_protob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tink.proto.aes_ctr_hmac_aead_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\034com.google.crypto.tink.protoP\001ZBgithub.com/tink-crypto/tink-go/v2/proto/aes_ctr_hmac_aead_go_proto'
  _globals['_AESCTRHMACAEADKEYFORMAT']._serialized_start=108
  _globals['_AESCTRHMACAEADKEYFORMAT']._serialized_end=258
  _globals['_AESCTRHMACAEADKEY']._serialized_start=261
  _globals['_AESCTRHMACAEADKEY']._serialized_end=396
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: tink/proto/config.proto
# Protobuf Python Version: 6.30.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    30,
    2,
    '',
    'tink/proto/config.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17tink/proto/config.proto\x12\x12google.crypto.tink\"\x8a\x01\n\x0cKeyTypeEntry\x12\x16\n\x0eprimitive_name\x18\x01 \x01(\t\x12\x10\n\x08type_url\x18\x02 \x01(\t\x12\x1b\n\x13key_manager_version\x18\x03 \x01(\r\x12\x17\n\x0fnew_key_allowed\x18\x04 \x01(\x08\x12\x16\n\x0e\x63\x61talogue_name\x18\x05 \x01(\t:\x02\x18\x01\"Z\n\x0eRegistryConfig\x12\x13\n\x0b\x63onfig_name\x18\x01 \x01(\t\x12/\n\x05\x65ntry\x18\x02 \x03(\x0b\x32 .google.crypto.tink.KeyTypeEntry:\x02\x18\x01\x42Y\n\x1c\x63om.google.crypto.tink.protoP\x01Z7github.com/tink-crypto/tink-go/v2/proto/config_go_protob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tink.proto.config_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\034com.google.crypto.tink.protoP\001Z7github.com/tink-crypto/tink-go/v2/proto/config_go_proto'
  _globals['_KEYTYPEENTRY']._loaded_options = None
  _globals['_KEYTYPEENTRY']._serialized_options = b'\030\001'
  _globals['_REGISTRYCONFIG']._loaded_options = None
  _globals['_REGISTRYCONFIG']._serialized_options = b'\030\001'
  _globals['_KEYTYPEENTRY']._serialized_start=48
  _globals['_KEYTYPEENTRY']._serialized_end=186
  _globals['_REGISTRYCONFIG']._serialized_start=188
  _globals['_REGISTRYCONFIG']._serialized_end=278
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: tink/proto/jwt_rsa_ssa_pkcs1.proto
# Protobuf Python Version: 6.30.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    30,
    2,
    '',
    'tink/proto/jwt_rsa_ssa_pkcs1.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\"tink/proto/jwt_rsa_ssa_pkcs1.proto\x12\x12google.crypto.tink\"\xe7\x01\n\x17JwtRsaSsaPkcs1PublicKey\x12\x0f\n\x07version\x18\x01 \x01(\r\x12>\n\talgorithm\x18\x02 \x01(\x0e\x32+.google.crypto.tink.JwtRsaSsaPkcs1Algorithm\x12\t\n\x01n\x18\x03 \x01(\x0c\x12\t\n\x01\x65\x18\x04 \x01(\x0c\x12I\n\ncustom_kid\x18\x05 \x01(\x0b\x32\x35.google.crypto.tink.JwtRsaSsaPkcs1PublicKey.CustomKid\x1a\x1a\n\tCustomKid\x12\r\n\x05value\x18\x01 \x01(\t\"\xb2\x01\n\x18JwtRsaSsaPkcs1PrivateKey\x12\x0f\n\x07version\x18\x01 \x01(\r\x12?\n\npublic_key\x18\x02 \x01(\x0b\x32+.google.crypto.tink.JwtRsaSsaPkcs1PublicKey\x12\t\n\x01\x64\x18\x03 \x01(\x0c\x12\t\n\x01p\x18\x04 \x01(\x0c\x12\t\n\x01q\x18\x05 \x01(\x0c\x12\n\n\x02\x64p\x18\x06 \x01(\x0c\x12\n\n\x02\x64q\x18\x07 \x01(\x0c\x12\x0b\n\x03\x63rt\x18\x08 \x01(\x0c\"\xa1\x01\n\x17JwtRsaSsaPkcs1KeyFormat\x12\x0f\n\x07version\x18\x01 \x01(\r\x12>\n\talgorithm\x18\x02 \x01(\x0e\x32+.google.crypto.tink.JwtRsaSsaPkcs1Algorithm\x12\x1c\n\x14modulus_size_in_bits\x18\x03 \x01(\r\x12\x17\n\x0fpublic_exponent\x18\x04 \x01(\x0c*J\n\x17JwtRsaSsaPkcs1Algorithm\x12\x0e\n\nRS_UNKNOWN\x10\x00\x12\t\n\x05RS256\x10\x01\x12\t\n\x05RS384\x10\x02\x12\t\n\x05RS512\x10\x03\x42`\n\x1c\x63om.google.crypto.tink.protoP\x01Z>github.com/tink-crypto/tink-go/v2/proto/rsa_ssa_pkcs1_go_protob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tink.proto.jwt_rsa_ssa_pkcs1_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\034com.google.crypto.tink.protoP\001Z>github.com/tink-crypto/tink-go/v2/proto/rsa_ssa_pkcs1_go_proto'
  _globals['_JWTRSASSAPKCS1ALGORITHM']._serialized_start=637
  _globals['_JWTRSASSAPKCS1ALGORITHM']._serialized_end=711
  _globals['_JWTRSASSAPKCS1PUBLICKEY']._serialized_start=59
  _globals['_JWTRSASSAPKCS1PUBLICKEY']._serialized_end=290
  _globals['_JWTRSASSAPKCS1PUBLICKEY_CUSTOMKID']._serialized_start=264
  _globals['_JWTRSASSAPKCS1PUBLICKEY_CUSTOMKID']._serialized_end=290
  _globals['_JWTRSASSAPKCS1PRIVATEKEY']._serialized_start=293
  _globals['_JWTRSASSAPKCS1PRIVATEKEY']._serialized_end=471
  _globals['_JWTRSASSAPKCS1KEYFORMAT']._serialized_start=474
  _globals['_JWTRSASSAPKCS1KEYFORMAT']._serialized_end=635
# @@protoc_insertion_point(module_scope)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Python wrapper of the wrapped C++ Streaming AEAD key manager."""

import io
from typing import BinaryIO

from tink import core
from tink.cc.pybind import tink_bindings
from tink.streaming_aead import _decrypting_stream
from tink.streaming_aead import _encrypting_stream
from tink.streaming_aead import _raw_streaming_aead
from tink.streaming_aead import _streaming_aead_wrapper


class _StreamingAeadCcToPyWrapper(_raw_streaming_aead.RawStreamingAead):
  """Transforms C++ StreamingAead into a RawStreamingAead Python primitive."""

  def __init__(self, cc_streaming_aead: tink_bindings.StreamingAead):
    self._cc_streaming_aead = cc_streaming_aead

  def new_raw_encrypting_stream(self, ciphertext_destination: BinaryIO,
                                associated_data: bytes) -> io.RawIOBase:
    return _encrypting_stream.RawEncryptingStream(self._cc_streaming_aead,
                                                  ciphertext_destination,
                                                  associated_data)

  def new_raw_decrypting_stream(
      self,
      ciphertext_source: BinaryIO,
      associated_data: bytes,
      close_ciphertext_source: bool) -> io.RawIOBase:
    return _decrypting_stream.RawDecryptingStream(
        self._cc_streaming_aead,
        ciphertext_source,
        associated_data,
        close_ciphertext_source=close_ciphertext_source)


def from_cc_registry(
    type_url: str) -> core.KeyManager[_raw_streaming_aead.RawStreamingAead]:
  return core.KeyManagerCcToPyWrapper(
      tink_bindings.StreamingAeadKeyManager.from_cc_registry(type_url),
      _raw_streaming_aead.RawStreamingAead, _StreamingAeadCcToPyWrapper)


def register() -> None:
  """Registers Streaming AEAD key managers and the wrapper in the Registry."""
  tink_bindings.register()
  for ident in (
      'AesCtrHmacStreamingKey',
      'AesGcmHkdfStreamingKey',
  ):
    type_url = 'type.googleapis.com/google.crypto.tink.{}'.format(ident)
    key_manager = core.KeyManagerCcToPyWrapper(
        tink_bindings.StreamingAeadKeyManager.from_cc_registry(type_url),
        _raw_streaming_aead.RawStreamingAead, _StreamingAeadCcToPyWrapper)
    core.Registry.register_key_manager(key_manager, new_key_allowed=True)
  core.Registry.register_primitive_wrapper(
      _streaming_aead_wrapper.StreamingAeadWrapper())



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Defines an Invalid JWT Error."""

import json
from typing import Any

from tink.jwt import _jwt_error


def json_dumps(json_data: Any) -> str:
  return json.dumps(json_data, separators=(',', ':'))


def _validate_str(value: str) -> None:
  """Uses encode('utf8') to check if value is a valid string."""
  _ = value.encode('utf8')


def _validate_value(value: Any) -> None:
  """Validates strings in a JSON object value."""
  # We don't need to check strings inside dicts, because json.loads will call
  # _dict_with_validation for these.
  if isinstance(value, str):
    _validate_str(value)
  if isinstance(value, list):
    for item in value:
      if isinstance(item, str):
        _validate_str(item)


def _dict_with_validation(pairs):
  """Validates pairs and returns a dict."""
  keys = set()
  for key, value in pairs:
    _validate_str(key)
    if key in keys:
      raise _jwt_error.JwtInvalidError(
          'Failed to parse JSON string, duplicated key')
    keys.add(key)
    _validate_value(value)
  return dict(pairs)


def json_loads(json_text: str) -> Any:
  """Does the same as json.loads, but with some additional validation."""
  try:
    return json.loads(json_text, object_pairs_hook=_dict_with_validation)
  except json.decoder.JSONDecodeError:
    raise _jwt_error.JwtInvalidError('Failed to parse JSON string')
  except RecursionError:
    raise _jwt_error.JwtInvalidError(
        'Failed to parse JSON string, too many recursions')
  except UnicodeEncodeError:
    raise _jwt_error.JwtInvalidError('invalid character')



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module defines the 'raw' interface for Streaming AEAD."""

import abc
import io
from typing import BinaryIO


class RawStreamingAead(metaclass=abc.ABCMeta):
  """Raw interface for streaming authenticated encryption with associated data.

  Streaming encryption is typically used for encrypting large plaintexts such
  as large files. This interface supports a streaming interface for symmetric
  encryption with authentication. The underlying encryption modes are selected
  so that partial plaintext can be obtained fast by decrypting and
  authenticating just a part of the ciphertext.

  This is the 'raw' interface implemented by key managers, based on
  io.RawIOBase. Users usually don't use this directly. Instead, they should use
  streaming_aead.StreamingAead, generated from a keyset_handle.
  """

  @abc.abstractmethod
  def new_raw_encrypting_stream(self, ciphertext_destination: BinaryIO,
                                associated_data: bytes) -> io.RawIOBase:
    """Returns a raw encrypting stream that writes to ciphertext_destination.

    The returned stream implements a writable io.RawIOBase interface. Users
    usually don't use this directly, they should use
    streaming_aead.StreamingAead instead.

    The ciphertext_destination's write() method is expected to present one of
    the following three behaviours in the case of a partial or failed write():
      - return a non-negative integer number of bytes written
      - return None (equivalent to returning 0)
      - raise BlockingIOError with characters_written set correctly to a
        non-negative integer (equivalent to returning that integer)
    In the case of a full write, the number of bytes written should be returned.
    The standard io.BufferedIOBase and io.RawIOBase base classes exhibit these
    behaviours and are hence supported.

    Args:
      ciphertext_destination: A writable binary file object to which ciphertext
        will be written. It must support write(), close(), closed, and
        writable().
      associated_data: Associated data to be used by the AEAD encryption. It is
        not included in the ciphertext and must be passed in as a parameter for
        decryption.

    Returns:
      A writable implementation of the io.RawIOBase interface that wraps around
      'ciphertext_destination', such that any bytes written to the wrapper are
      AEAD-encrypted using 'associated_data' as associated authenticated data.
      Closing this wrapper also closes the ciphertext_source.
    Raises:
      tink.TinkError if the creation fails.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def new_raw_decrypting_stream(
      self,
      ciphertext_source: BinaryIO,
      associated_data: bytes,
      close_ciphertext_source: bool) -> io.RawIOBase:
    """Returns a raw decrypting stream that reads from ciphertext_source.

    The returned stream implements a readable io.RawIOBase interface. Users
    usually don't use this directly, they should use
    streaming_aead.StreamingAead instead.

    The cipertext_source's read() method is expected to return an empty bytes
    object if the stream is already at EOF. In the case where the stream is not
    at EOF yet but no data is available at the moment, it is expected to either
    block until data is available, return None or raise a BlockingIOError.
    The standard io.BufferedIOBase and io.RawIOBase base classes exhibit these
    behaviours and are hence supported.

    Args:
      ciphertext_source: A readable binary file object from which ciphertext
        will be read.
      associated_data: Associated data to be used by the AEAD decryption. It
        must match the associated_data supplied for the encryption.
      close_ciphertext_source: Whether ciphertext_source should be closed when
      close() is called.

    Returns:
      A readable implementation of the io.RawIOBase interface that wraps around
      'ciphertext_source', such that any bytes read from the wrapper are
      AEAD-decrypted using 'associated_data' as associated authenticated data.
    Raises:
      tink.TinkError if the creation fails.
    """
    raise NotImplementedError()
